"""Tool management for LLMling agents."""

from __future__ import annotations

import asyncio
from contextlib import AsyncExitStack
from typing import TYPE_CHECKING, Any, Self

from llmling_agent.log import get_logger
from llmling_agent.resource_providers import ResourceProvider
from llmling_agent_config.mcp_server import BaseMCPServerConfig
from llmling_agent_config.resources import ResourceInfo


if TYPE_CHECKING:
    from typing import Literal

    from fastmcp.client.sampling import ClientSamplingHandler

    from llmling_agent.prompts.prompts import MCPClientPrompt
    from llmling_agent.tools.base import Tool
    from llmling_agent_config.mcp_server import MCPServerConfig


logger = get_logger(__name__)


class MCPResourceProvider(ResourceProvider):
    """Resource provider for a single MCP server."""

    def __init__(
        self,
        server: MCPServerConfig | str,
        name: str = "mcp",
        owner: str | None = None,
        source: Literal["pool", "node"] = "node",
        sampling_callback: ClientSamplingHandler[Any] | None = None,
        accessible_roots: list[str] | None = None,
    ) -> None:
        from llmling_agent.mcp_server import MCPClient

        super().__init__(name, owner=owner)
        self.server = BaseMCPServerConfig.from_string(server) if isinstance(server, str) else server
        self.source = source
        self.exit_stack = AsyncExitStack()
        self._accessible_roots = accessible_roots
        self._sampling_callback = sampling_callback

        # Tool caching
        self._tools_cache: list[Tool] | None = None
        self._saved_enabled_states: dict[str, bool] = {}

        # Prompt caching
        self._prompts_cache: list[MCPClientPrompt] | None = None

        # Resource caching
        self._resources_cache: list[ResourceInfo] | None = None

        self.client = MCPClient(
            config=self.server,
            sampling_callback=self._sampling_callback,
            accessible_roots=self._accessible_roots,
            tool_change_callback=self._on_tools_changed,
            prompt_change_callback=self._on_prompts_changed,
            resource_change_callback=self._on_resources_changed,
        )

    def __repr__(self) -> str:
        return f"MCPResourceProvider({self.server!r}, source={self.source!r})"

    async def __aenter__(self) -> Self:
        try:
            await self.exit_stack.enter_async_context(self.client)
        except Exception as e:
            # Clean up in case of error
            await self.__aexit__(type(e), e, e.__traceback__)
            msg = "Failed to initialize MCP manager"
            raise RuntimeError(msg) from e

        return self

    async def __aexit__(self, *args: object) -> None:
        try:
            try:
                # Clean up exit stack (which includes MCP clients)
                await self.exit_stack.aclose()
            except RuntimeError as e:
                if "different task" in str(e):
                    # Handle task context mismatch
                    current_task = asyncio.current_task()
                    if current_task:
                        loop = asyncio.get_running_loop()
                        await loop.create_task(self.exit_stack.aclose())
                else:
                    raise

        except Exception as e:
            msg = "Error during MCP manager cleanup"
            logger.exception(msg, exc_info=e)
            raise RuntimeError(msg) from e

    async def _on_tools_changed(self) -> None:
        """Callback when tools change on the MCP server."""
        logger.info("MCP tool list changed, refreshing provider cache")
        self._saved_enabled_states = {t.name: t.enabled for t in self._tools_cache or []}
        self._tools_cache = None

    async def _on_prompts_changed(self) -> None:
        """Callback when prompts change on the MCP server."""
        logger.info("MCP prompt list changed, refreshing provider cache")
        self._prompts_cache = None

    async def _on_resources_changed(self) -> None:
        """Callback when resources change on the MCP server."""
        logger.info("MCP resource list changed, refreshing provider cache")
        self._resources_cache = None

    async def refresh_tools_cache(self) -> None:
        """Refresh the tools cache by fetching from client."""
        try:
            # Get fresh tools from client
            mcp_tools = await self.client.list_tools()
            all_tools: list[Tool] = []

            for tool in mcp_tools:
                try:
                    tool_info = self.client.convert_tool(tool)
                    all_tools.append(tool_info)
                except Exception:
                    logger.exception("Failed to create MCP tool", name=tool.name)
                    continue

            # Restore enabled states from saved states
            for tool_info in all_tools:
                if tool_info.name in self._saved_enabled_states:
                    tool_info.enabled = self._saved_enabled_states[tool_info.name]

            self._tools_cache = all_tools
            logger.debug("Refreshed MCP tools cache", num_tools=len(all_tools))
        except Exception:
            logger.exception("Failed to refresh MCP tools cache")
            self._tools_cache = []

    async def get_tools(self) -> list[Tool]:
        """Get cached tools, refreshing if necessary."""
        if self._tools_cache is None:
            await self.refresh_tools_cache()

        return self._tools_cache or []

    async def refresh_prompts_cache(self) -> None:
        """Refresh the prompts cache by fetching from client."""
        from llmling_agent.prompts.prompts import MCPClientPrompt

        try:
            result = await self.client.list_prompts()
            all_prompts: list[MCPClientPrompt] = []

            for prompt in result:
                try:
                    converted = MCPClientPrompt.from_fastmcp(self.client, prompt)
                    all_prompts.append(converted)
                except Exception:
                    logger.exception("Failed to convert prompt", name=prompt.name)
                    continue

            self._prompts_cache = all_prompts
            logger.debug("Refreshed MCP prompts cache", num_prompts=len(all_prompts))
        except Exception:
            logger.exception("Failed to refresh MCP prompts cache")
            self._prompts_cache = []

    async def get_prompts(self) -> list[MCPClientPrompt]:  # type: ignore
        """Get cached prompts, refreshing if necessary."""
        if self._prompts_cache is None:
            await self.refresh_prompts_cache()

        return self._prompts_cache or []

    async def refresh_resources_cache(self) -> None:
        """Refresh the resources cache by fetching from client."""
        try:
            result = await self.client.list_resources()
            all_resources: list[ResourceInfo] = []

            for resource in result:
                try:
                    converted = await ResourceInfo.from_mcp_resource(resource)
                    all_resources.append(converted)
                except Exception:
                    logger.exception("Failed to convert resource", name=resource.name)
                    continue

            self._resources_cache = all_resources
            logger.debug("Refreshed MCP resources cache", num_resources=len(all_resources))
        except Exception:
            logger.exception("Failed to refresh MCP resources cache")
            self._resources_cache = []

    async def get_resources(self) -> list[ResourceInfo]:
        """Get cached resources, refreshing if necessary."""
        if self._resources_cache is None:
            await self.refresh_resources_cache()

        return self._resources_cache or []


if __name__ == "__main__":
    from llmling_agent_config.mcp_server import StdioMCPServerConfig

    cfg = StdioMCPServerConfig(
        command="uv",
        args=["run", "/home/phil65/dev/oss/llmling-agent/tests/mcp_server/server.py"],
    )

    async def main() -> None:
        manager = MCPResourceProvider(cfg)
        async with manager:
            prompts = await manager.get_prompts()
            print(f"Found prompts: {prompts}")

            # Test static prompt (no arguments)
            static_prompt = next(p for p in prompts if p.name == "static_prompt")
            print(f"\n--- Testing static prompt: {static_prompt} ---")
            components = await static_prompt.get_components()
            assert components, "No prompt components found"
            print(f"Found {len(components)} prompt components:")
            for i, component in enumerate(components):
                comp_type = type(component).__name__
                print(f"  {i + 1}. {comp_type}: {component.content}")

            # Test dynamic prompt (with arguments)
            dynamic_prompt = next(p for p in prompts if p.name == "dynamic_prompt")
            print(f"\n--- Testing dynamic prompt: {dynamic_prompt} ---")
            components = await dynamic_prompt.get_components(
                arguments={"some_arg": "Hello, world!"}
            )
            assert components, "No prompt components found"
            print(f"Found {len(components)} prompt components:")
            for i, component in enumerate(components):
                comp_type = type(component).__name__
                print(f"  {i + 1}. {comp_type}: {component.content}")

    asyncio.run(main())



"""MCP server integration for LLMling agent."""

from llmling_agent.mcp_server.client import MCPClient
from llmling_agent.mcp_server.tool_bridge import (
    BridgeConfig,
    ToolBridgeRegistry,
    ToolManagerBridge,
    create_tool_bridge,
)

__all__ = [
    "BridgeConfig",
    "MCPClient",
    "ToolBridgeRegistry",
    "ToolManagerBridge",
    "create_tool_bridge",
]



"""Models for agent configuration."""

from __future__ import annotations

from collections.abc import Sequence  # noqa: TC003
from pathlib import Path
import re
from typing import TYPE_CHECKING, Any, Literal, assert_never
from uuid import UUID

from llmling_models.configs import AnyModelConfig  # noqa: TC002
from pydantic import ConfigDict, Field, model_validator
from pydantic_ai import UsageLimits  # noqa: TC002
from schemez import InlineSchemaDef
from tokonomics.model_names import ModelName  # noqa: TC002
from toprompt import render_prompt
from upathtools import to_upath

from llmling_agent import log
from llmling_agent.common_types import EndStrategy  # noqa: TC001
from llmling_agent.prompts.prompts import PromptMessage, StaticPrompt
from llmling_agent.resource_providers import StaticResourceProvider
from llmling_agent.utils.importing import import_class
from llmling_agent_config.knowledge import Knowledge  # noqa: TC001
from llmling_agent_config.nodes import NodeConfig
from llmling_agent_config.output_types import StructuredResponseConfig  # noqa: TC001
from llmling_agent_config.session import MemoryConfig, SessionQuery
from llmling_agent_config.system_prompts import PromptConfig  # noqa: TC001
from llmling_agent_config.tools import BaseToolConfig, ToolConfig  # noqa: TC001
from llmling_agent_config.toolsets import ToolsetConfig  # noqa: TC001
from llmling_agent_config.workers import WorkerConfig  # noqa: TC001


if TYPE_CHECKING:
    from llmling_agent.prompts.prompts import BasePrompt
    from llmling_agent.resource_providers import ResourceProvider
    from llmling_agent.tools.base import Tool


ToolConfirmationMode = Literal["always", "never", "per_tool"]
ToolMode = Literal["codemode"]

logger = log.get_logger(__name__)


# Claude Code model alias mapping
CLAUDE_MODEL_ALIASES: dict[str, str] = {
    "sonnet": "anthropic:claude-sonnet-4-20250514",
    "opus": "anthropic:claude-opus-4-20250514",
    "haiku": "anthropic:claude-haiku-3-5-20241022",
}

# Claude Code permissionMode to ToolConfirmationMode mapping
PERMISSION_MODE_MAP: dict[str, ToolConfirmationMode] = {
    "default": "per_tool",
    "acceptEdits": "never",
    "bypassPermissions": "never",
    # 'plan' and 'ignore' don't map well, default to per_tool
}


def parse_agent_file(file_path: str, *, skills_registry: Any | None = None) -> AgentConfig:  # noqa: PLR0915
    """Parse agent markdown file to AgentConfig.

    Supports both Claude Code and OpenCode formats with auto-detection.
    Also supports local and remote paths via UPath.

    Format documentation:
    - Claude Code: https://code.claude.com/docs/en/sub-agents.md
    - OpenCode: https://raw.githubusercontent.com/sst/opencode/refs/heads/dev/packages/web/src/content/docs/agents.mdx

    Args:
        file_path: Path to .md file with YAML frontmatter (local or remote)
        skills_registry: Optional skills registry for loading skills

    Claude Code format:
        ```markdown
        ---
        name: agent-name
        description: When to use this agent
        tools: tool1, tool2  # comma-separated
        model: sonnet  # sonnet/opus/haiku/inherit
        permissionMode: default
        skills: skill1, skill2
        ---

        System prompt content.
        ```

    OpenCode format:
        ```markdown
        ---
        description: Agent description
        mode: subagent  # primary/subagent/all
        model: anthropic/claude-sonnet-4-20250514
        temperature: 0.1
        maxSteps: 5
        tools:
          write: false
          edit: false
        permission:
          edit: deny
          bash:
            "git diff": allow
        ---

        System prompt content.
        ```

    Additional llmling-agent fields in frontmatter are also supported.
    """
    import yamling

    path = to_upath(file_path)
    content = path.read_text("utf-8")
    # Extract YAML frontmatter
    frontmatter_match = re.match(r"^---\s*\n(.*?)\n---\s*\n?", content, re.DOTALL)
    if not frontmatter_match:
        msg = f"No YAML frontmatter found in {file_path}"
        raise ValueError(msg)

    try:
        metadata = yamling.load_yaml(frontmatter_match.group(1))
    except yamling.YAMLError as e:
        msg = f"Invalid YAML frontmatter in {file_path}: {e}"
        raise ValueError(msg) from e

    if not isinstance(metadata, dict):
        msg = f"YAML frontmatter must be a dictionary in {file_path}"
        raise ValueError(msg)  # noqa: TRY004

    # Extract system prompt (everything after frontmatter)
    system_prompt = content[frontmatter_match.end() :].strip()
    # Build AgentConfig kwargs
    config_kwargs: dict[str, Any] = {}
    # Required field
    if description := metadata.get("description"):
        config_kwargs["description"] = description
    # Detect format and parse accordingly
    is_opencode = (
        any(key in metadata for key in ["mode", "temperature", "maxSteps", "disable"])
        or ("tools" in metadata and isinstance(metadata["tools"], dict))
        or ("permission" in metadata and isinstance(metadata["permission"], dict))
    )
    # Model handling
    if model := metadata.get("model"):
        if model == "inherit":
            pass  # Leave as None, will use default
        elif model in CLAUDE_MODEL_ALIASES:
            config_kwargs["model"] = CLAUDE_MODEL_ALIASES[model]
        else:
            config_kwargs["model"] = model

    # OpenCode-specific fields
    if is_opencode:
        # Temperature
        if temperature := metadata.get("temperature"):
            logger.debug(
                "OpenCode temperature %r in %s (not directly supported)", temperature, file_path
            )

        # MaxSteps
        if max_steps := metadata.get("maxSteps"):
            logger.debug(
                "OpenCode maxSteps %r in %s (not directly supported)", max_steps, file_path
            )

        # Disable
        if disable := metadata.get("disable"):
            logger.debug("OpenCode disable %r in %s (not directly supported)", disable, file_path)

        # Mode (primary/subagent/all) - we don't have direct equivalent
        if mode := metadata.get("mode"):
            logger.debug("OpenCode mode %r in %s (informational only)", mode, file_path)

        # Permission handling (granular per-tool)
        if permission := metadata.get("permission"):
            # For now, just check if edit permissions are restrictive
            edit_perm = permission.get("edit") if isinstance(permission, dict) else None
            if edit_perm in ("deny", "ask"):
                config_kwargs["requires_tool_confirmation"] = (
                    "always" if edit_perm == "ask" else "never"
                )
            logger.debug("OpenCode permission %r in %s (partial mapping)", permission, file_path)

        # TODO: OpenCode tools dict format (tool_name: bool) - needs toolset integration
        if (tools := metadata.get("tools")) and isinstance(tools, dict):
            logger.debug("OpenCode tools dict %r in %s (not yet supported)", tools, file_path)

    # Claude Code-specific fields
    else:
        # Permission mode mapping
        if permission_mode := metadata.get("permissionMode"):
            if mapped := PERMISSION_MODE_MAP.get(permission_mode):
                config_kwargs["requires_tool_confirmation"] = mapped
            else:
                logger.warning(
                    "Unknown permissionMode %r in %s, using default",
                    permission_mode,
                    file_path,
                )

        # TODO: Claude Code tools string format (comma-separated) - needs toolset integration
        if (tools := metadata.get("tools")) and isinstance(tools, str):
            logger.debug("Claude Code tools string %r in %s (not yet supported)", tools, file_path)

        # Skills handling (Claude Code only)
        if (skills_str := metadata.get("skills")) and skills_registry is not None:
            skill_names = [s.strip() for s in skills_str.split(",")]
            for skill_name in skill_names:
                if skill_name not in skills_registry:
                    logger.warning(
                        "Skill %r from %s not found in registry, ignoring",
                        skill_name,
                        file_path,
                    )

    # System prompt from markdown body
    if system_prompt:
        config_kwargs["system_prompts"] = [system_prompt]

    # Pass through any additional llmling-agent specific fields
    passthrough_fields = {
        "inherits",
        "toolsets",
        "session",
        "output_type",
        "retries",
        "output_retries",
        "end_strategy",
        "avatar",
        "config_file_path",
        "knowledge",
        "workers",
        "debug",
        "usage_limits",
        "tool_mode",
        "display_name",
        "triggers",
    }
    for field in passthrough_fields:
        if field in metadata:
            config_kwargs[field] = metadata[field]

    return AgentConfig(**config_kwargs)


class AgentConfig(NodeConfig):
    """Configuration for a single agent in the system.

    Defines an agent's complete configuration including its model, environment,
    and behavior settings.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/agent_configuration/
    """

    model_config = ConfigDict(
        json_schema_extra={
            "x-icon": "octicon:hubot-16",
            "x-doc-title": "Agent Configuration",
        }
    )

    inherits: str | None = Field(default=None, title="Inheritance source")
    """Name of agent config to inherit from"""

    model: AnyModelConfig | ModelName | str | None = Field(
        default=None,
        examples=["openai:gpt-5-nano"],
        title="Model configuration or name",
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/model_configuration/"
        },
    )
    """The model to use for this agent. Can be either a simple model name
    string (e.g. 'openai:gpt-5') or a structured model definition.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/model_configuration/
    """

    tools: list[ToolConfig | str] = Field(
        default_factory=list,
        examples=[
            ["webbrowser:open", "builtins:print"],
            [
                {
                    "type": "import",
                    "import_path": "webbrowser:open",
                    "name": "web_browser",
                }
            ],
        ],
        title="Tool configurations",
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/tool_configuration/"
        },
    )
    """A list of tools to register with this agent.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/tool_configuration/
    """

    toolsets: list[ToolsetConfig] = Field(
        default_factory=list,
        examples=[
            [
                {
                    "type": "openapi",
                    "spec": "https://api.example.com/openapi.json",
                    "namespace": "api",
                },
                {
                    "type": "file_access",
                },
                {
                    "type": "composio",
                    "user_id": "user123@example.com",
                    "toolsets": ["github", "slack"],
                },
            ],
        ],
        title="Toolset configurations",
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/toolset_configuration/"
        },
    )
    """Toolset configurations for extensible tool collections.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/toolset_configuration/
    """

    session: str | SessionQuery | MemoryConfig | None = Field(
        default=None,
        examples=["main_session", "user_123"],
        title="Session configuration",
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/session_configuration/"
        },
    )
    """Session configuration for conversation recovery.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/session_configuration/
    """

    output_type: str | StructuredResponseConfig | None = Field(
        default=None,
        examples=["json_response", "code_output"],
        title="Response type",
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/response_configuration/"
        },
    )
    """Name of the response definition to use.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/response_configuration/
    """

    retries: int = Field(default=1, ge=0, examples=[1, 3], title="Model retries")
    """Number of retries for failed operations (maps to pydantic-ai's retries)"""

    output_retries: int | None = Field(default=None, examples=[1, 3], title="Output retries")
    """Max retries for result validation"""

    end_strategy: EndStrategy = Field(
        default="early",
        examples=["early", "exhaust"],
        title="Tool execution strategy",
    )
    """The strategy for handling multiple tool calls when a final result is found"""

    avatar: str | None = Field(
        default=None,
        examples=["https://example.com/avatar.png", "/assets/robot.jpg"],
        title="Avatar image",
    )
    """URL or path to agent's avatar image"""

    system_prompts: Sequence[str | PromptConfig] = Field(
        default_factory=list,
        title="System prompts",
        examples=[["You are an AI assistant."]],
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/system_prompts_configuration/"
        },
    )
    """System prompts for the agent. Can be strings or structured prompt configs.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/system_prompts_configuration/
    """

    # context_sources: list[ContextSource] = Field(default_factory=list)
    # """Initial context sources to load"""

    knowledge: Knowledge | None = Field(
        default=None,
        title="Knowledge sources",
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/knowledge_configuration/"
        },
    )
    """Knowledge sources for this agent.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/knowledge_configuration/
    """

    workers: list[WorkerConfig] = Field(
        default_factory=list,
        examples=[
            [{"type": "agent", "name": "web_agent", "reset_history_on_run": True}],
            [{"type": "team", "name": "analysis_team"}],
        ],
        title="Worker agents",
        json_schema_extra={
            "documentation_url": "https://phil65.github.io/llmling-agent/YAML%20Configuration/worker_configuration/"
        },
    )
    """Worker agents which will be available as tools.

    Docs: https://phil65.github.io/llmling-agent/YAML%20Configuration/worker_configuration/
    """

    requires_tool_confirmation: ToolConfirmationMode = Field(
        default="per_tool",
        examples=["always", "never", "per_tool"],
        title="Tool confirmation mode",
    )
    """How to handle tool confirmation:
    - "always": Always require confirmation for all tools
    - "never": Never require confirmation (ignore tool settings)
    - "per_tool": Use individual tool settings
    """

    debug: bool = Field(default=False, title="Debug mode")
    """Enable debug output for this agent."""

    usage_limits: UsageLimits | None = Field(default=None, title="Usage limits")
    """Usage limits for this agent."""

    tool_mode: ToolMode | None = Field(
        default=None,
        examples=["codemode"],
        title="Tool execution mode",
    )
    """Tool execution mode:
    - None: Default mode - tools are called directly
    - "codemode": Tools are wrapped in a Python execution environment
    """

    @model_validator(mode="before")
    @classmethod
    def validate_output_type(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Convert result type and apply its settings."""
        output_type = data.get("output_type")
        if isinstance(output_type, dict):
            # Extract response-specific settings
            tool_name = output_type.pop("result_tool_name", None)
            tool_description = output_type.pop("result_tool_description", None)
            retries = output_type.pop("output_retries", None)

            # Convert remaining dict to ResponseDefinition
            if "type" not in output_type["response_schema"]:
                output_type["response_schema"]["type"] = "inline"
            data["output_type"]["response_schema"] = InlineSchemaDef(**output_type)

            # Apply extracted settings to agent config
            if tool_name:
                data["result_tool_name"] = tool_name
            if tool_description:
                data["result_tool_description"] = tool_description
            if retries is not None:
                data["output_retries"] = retries

        return data

    @model_validator(mode="before")
    @classmethod
    def handle_model_types(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Convert model inputs to appropriate format."""
        if isinstance((model := data.get("model")), str):
            data["model"] = {"type": "string", "identifier": model}
        return data

    def get_toolsets(self) -> list[ResourceProvider]:
        """Get all resource providers for this agent."""
        providers: list[ResourceProvider] = []

        # Add providers from toolsets
        for toolset_config in self.toolsets:
            try:
                provider = toolset_config.get_provider()
                providers.append(provider)
            except Exception as e:
                msg = "Failed to create provider for toolset"
                logger.exception(msg, toolset_config)
                raise ValueError(msg) from e

        return providers

    def get_tool_provider(self) -> ResourceProvider | None:
        """Get tool provider for this agent."""
        from llmling_agent.tools.base import Tool

        # Create provider for static tools
        if not self.tools:
            return None
        static_tools: list[Tool] = []
        for tool_config in self.tools:
            try:
                match tool_config:
                    case str():
                        if tool_config.startswith("crewai_tools"):
                            obj = import_class(tool_config)()
                            static_tools.append(Tool.from_crewai_tool(obj))
                        elif tool_config.startswith("langchain"):
                            obj = import_class(tool_config)()
                            static_tools.append(Tool.from_langchain_tool(obj))
                        else:
                            tool = Tool.from_callable(tool_config)
                            static_tools.append(tool)
                    case BaseToolConfig():
                        static_tools.append(tool_config.get_tool())
            except Exception:
                logger.exception("Failed to load tool", config=tool_config)
                continue

        return StaticResourceProvider(name="builtin", tools=static_tools)

    def get_session_config(self) -> MemoryConfig:
        """Get resolved memory configuration."""
        match self.session:
            case str() | UUID():
                return MemoryConfig(session=SessionQuery(name=str(self.session)))
            case SessionQuery():
                return MemoryConfig(session=self.session)
            case MemoryConfig():
                return self.session
            case None:
                return MemoryConfig()
            case _ as unreachable:
                assert_never(unreachable)

    def get_system_prompts(self) -> list[BasePrompt]:
        """Get all system prompts as BasePrompts."""
        from llmling_agent_config.system_prompts import (
            FilePromptConfig,
            FunctionPromptConfig,
            LibraryPromptConfig,
            StaticPromptConfig,
        )

        prompts: list[BasePrompt] = []
        for prompt in self.system_prompts:
            match prompt:
                case (str() as content) | StaticPromptConfig(content=content):
                    # Convert string to StaticPrompt
                    msgs = [PromptMessage(role="system", content=content)]
                    static = StaticPrompt(name="system", description="System prompt", messages=msgs)
                    prompts.append(static)
                case FilePromptConfig(path=path):
                    template_path = Path(path)
                    if not template_path.is_absolute() and self.config_file_path:
                        base_path = Path(self.config_file_path).parent
                        template_path = base_path / path
                    template_content = template_path.read_text("utf-8")
                    # Create a template-based prompt (for now as StaticPrompt with placeholder)
                    static_prompt = StaticPrompt(
                        name="system",
                        description=f"File prompt: {path}",
                        messages=[PromptMessage(role="system", content=template_content)],
                    )
                    prompts.append(static_prompt)
                case LibraryPromptConfig(reference=ref):
                    # Create placeholder for library prompts (resolved by manifest)
                    msg = PromptMessage(role="system", content=f"[LIBRARY:{ref}]")
                    static = StaticPrompt(name="system", description=f"Ref: {ref}", messages=[msg])
                    prompts.append(static)
                case FunctionPromptConfig(arguments=arguments, function=function):
                    # Import and call the function to get prompt content
                    content = function(**arguments)
                    static_prompt = StaticPrompt(
                        name="system",
                        description=f"Function prompt: {function}",
                        messages=[PromptMessage(role="system", content=content)],
                    )
                    prompts.append(static_prompt)
                case _ as unreachable:
                    assert_never(unreachable)
        return prompts

    def render_system_prompts(self, context: dict[str, Any] | None = None) -> list[str]:
        """Render system prompts with context."""
        from llmling_agent_config.system_prompts import (
            FilePromptConfig,
            FunctionPromptConfig,
            LibraryPromptConfig,
            StaticPromptConfig,
        )

        context = context or {"name": self.name, "id": 1, "model": self.model}
        rendered_prompts: list[str] = []
        for prompt in self.system_prompts:
            match prompt:
                case (str() as content) | StaticPromptConfig(content=content):
                    rendered_prompts.append(render_prompt(content, {"agent": context}))
                case FilePromptConfig(path=path, variables=variables):
                    # Load and render Jinja template from file
                    template_path = Path(path)
                    if not template_path.is_absolute() and self.config_file_path:
                        base_path = Path(self.config_file_path).parent
                        template_path = base_path / path

                    template_content = template_path.read_text("utf-8")
                    template_ctx = {"agent": context, **variables}
                    rendered_prompts.append(render_prompt(template_content, template_ctx))
                case LibraryPromptConfig(reference=reference):
                    # This will be handled by the manifest's get_agent method
                    # For now, just add a placeholder
                    rendered_prompts.append(f"[LIBRARY:{reference}]")
                case FunctionPromptConfig(function=function, arguments=arguments):
                    # Import and call the function to get prompt content
                    content = function(**arguments)
                    rendered_prompts.append(render_prompt(content, {"agent": context}))

        return rendered_prompts


if __name__ == "__main__":
    model = "openai:gpt-5-nano"
    agent_cfg = AgentConfig(name="test_agent", model=model, tools=["crewai_tools.BraveSearchTool"])
    print(agent_cfg)



"""Task configuration."""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from pydantic import ConfigDict, Field, ImportString
from schemez import Schema

from llmling_agent.prompts.prompts import BasePrompt
from llmling_agent.tools.base import Tool
from llmling_agent_config.knowledge import Knowledge
from llmling_agent_config.tools import ImportToolConfig


if TYPE_CHECKING:
    from llmling_agent.agent import Agent


class Job[TDeps, TResult = str](Schema):
    """A task is a piece of work that can be executed by an agent.

    Requirements:
    - The agent must have compatible dependencies (required_dependency)
    - The agent must produce the specified result type (required_return_type)

    Equipment:
    - The task provides necessary tools for execution (tools)
    - Tools are temporarily available during task execution
    """

    name: str | None = Field(
        default=None,
        examples=["analyze_code", "generate_summary", "translate_text"],
        title="Task name",
    )
    """Technical identifier (automatically set from config key during registration)"""

    description: str | None = Field(
        default=None,
        examples=[
            "Analyze code for issues",
            "Generate a summary",
            "Translate text to target language",
        ],
        title="Task description",
    )
    """Human-readable description of what this task does"""

    prompt: str | ImportString[str] | BasePrompt = Field(
        examples=["Analyze the following code", "llmling_agent.prompts.prompts:BasePrompt"],
        title="Task prompt",
    )
    """The task instruction/prompt."""

    required_return_type: ImportString[type[TResult]] = Field(
        default="str",
        validate_default=True,
        examples=["str", "dict", "builtins:str"],
        title="Required return type",
    )
    """Expected type of the task result."""

    required_dependency: ImportString[type[TDeps]] | None = Field(
        default=None,
        validate_default=True,
        examples=["llmling_agent.agent.context:AgentContext"],
        title="Required dependencies",
    )
    """Dependencies or context data needed for task execution"""

    requires_vision: bool = Field(default=False, title="Requires vision capabilities")
    """Whether the agent requires vision"""

    knowledge: Knowledge | None = Field(default=None, title="Task knowledge sources")
    """Optional knowledge sources for this task:
    - Simple file/URL paths
    - Rich resource definitions
    - Prompt templates
    """

    tools: list[ImportString[Callable[..., Any]] | ImportToolConfig] = Field(
        default_factory=list,
        examples=[
            [
                {
                    "type": "import",
                    "import_path": "webbrowser:open",
                    "name": "web_browser",
                }
            ],
            ["builtins:print"],
            [
                {
                    "type": "import",
                    "import_path": "builtins:print",
                    "requires_confirmation": True,
                }
            ],
        ],
        title="Task tools",
    )
    """Tools needed for this task."""

    min_context_tokens: int | None = Field(
        default=None,
        examples=[1000, 4000, 8000],
        title="Minimum context tokens",
    )
    """Minimum amount of required context size."""

    model_config = ConfigDict(frozen=True)

    async def can_be_executed_by(self, agent: Agent[Any, Any]) -> bool:
        """Check if agent meets all requirements for this task."""
        # Check dependencies
        if self.required_dependency and not issubclass(
            agent.deps_type or type(None), self.required_dependency
        ):
            return False

        if agent._output_type != self.required_return_type:
            return False

        # Check vision capabilities
        if self.requires_vision:
            from llmling_agent.utils.model_capabilities import supports_vision

            if not await supports_vision(agent.model_name):
                return False

        return True

    @property
    def tool_configs(self) -> list[ImportToolConfig]:
        """Get all tools as ToolConfig instances."""
        return [
            tool if isinstance(tool, ImportToolConfig) else ImportToolConfig(import_path=tool)
            for tool in self.tools
        ]

    async def get_prompt(self) -> str:
        if isinstance(self.prompt, BasePrompt):
            messages = await self.prompt.format()
            return "\n\n".join(m.get_text_content() for m in messages)
        return self.prompt

    def get_tools(self) -> list[Tool]:
        """Get all tools as Tool instances."""
        tools: list[Tool] = []
        for tool in self.tools:
            match tool:
                case str():
                    tools.append(Tool.from_callable(tool))
                case ImportToolConfig() as config:
                    tools.append(config.get_tool())
                case Tool():
                    tools.append(tool)
                case _:
                    raise ValueError(f"Invalid tool type: {type(tool)}")
        return tools



"""Claude Code Skills registry with auto-discovery."""

from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, ClassVar

from fsspec import AbstractFileSystem
from fsspec.asyn import AsyncFileSystem
from fsspec.implementations.asyn_wrapper import (
    AsyncFileSystemWrapper,
)
from upath import UPath
from upathtools.helpers import upath_to_fs

from llmling_agent.log import get_logger
from llmling_agent.skills.skill import Skill
from llmling_agent.tools.exceptions import ToolError
from llmling_agent.utils.baseregistry import BaseRegistry


if TYPE_CHECKING:
    from collections.abc import Sequence

    from upath.types import JoinablePathLike


SKILL_NAME_LIMIT = 64
SKILL_DESCRIPTION_LIMIT = 1024
logger = get_logger(__name__)


class SkillsRegistry(BaseRegistry[str, Skill]):
    """Registry for Claude Code Skills with auto-discovery."""

    DEFAULT_SKILL_PATHS: ClassVar = ["~/.claude/skills/", ".claude/skills/"]

    def __init__(self, skills_dirs: Sequence[JoinablePathLike] | None = None) -> None:
        """Initialize with custom skill directories or auto-detect."""
        super().__init__()
        if skills_dirs:
            self.skills_dirs = [UPath(i) for i in skills_dirs or []]
        else:
            self.skills_dirs = [UPath(i) for i in self.DEFAULT_SKILL_PATHS or []]

    async def discover_skills(self) -> None:
        """Scan filesystem and register all found skills.

        Args:
            filesystem: Optional async filesystem to use. If None, will use upath_to_fs()
                       to get appropriate filesystem for each skills directory.
        """
        for skills_dir in self.skills_dirs:
            await self.register_skills_from_path(skills_dir)

    async def register_skills_from_path(
        self,
        skills_dir: UPath | AbstractFileSystem,
        **storage_options: Any,
    ) -> None:
        """Register skills from a given path.

        Args:
            skills_dir: Path to the directory containing skills.
            storage_options: Additional options to pass to the filesystem.
        """
        if isinstance(skills_dir, AbstractFileSystem):
            fs = skills_dir
            if not isinstance(fs, AsyncFileSystem):
                fs = AsyncFileSystemWrapper(fs)
        else:
            fs = upath_to_fs(skills_dir, **storage_options)

        try:
            # List entries in skills directory
            entries = await fs._ls(fs.root_marker, detail=True)
        except FileNotFoundError:
            logger.warning("Skills directory not found", path=skills_dir)
            return
        # Filter for directories that might contain skills
        skill_dirs = [entry for entry in entries if await fs._isdir(entry["name"])]
        if not skill_dirs:
            logger.info("No skills found", skills_dir=skills_dir)
            return
        logger.info("Found skills", skills=skill_dirs, skills_dir=skills_dir)
        for skill_entry in skill_dirs:
            skill_name = skill_entry["name"].lstrip("./")
            skill_dir_path = skills_dir / skill_name
            try:
                await fs._cat(f"{skill_name}/SKILL.md")
            except FileNotFoundError:
                continue

            try:
                skill = self._parse_skill(skill_dir_path)
                self.register(skill.name, skill, replace=True)
            except Exception as e:  # noqa: BLE001
                # Log but don't fail discovery for one bad skill
                print(f"Warning: Failed to parse skill at {skill_dir_path}: {e}")

    def _parse_skill(self, skill_dir: JoinablePathLike) -> Skill:
        """Parse a SKILL.md file and extract metadata."""
        skill_file = UPath(skill_dir) / "SKILL.md"
        content = skill_file.read_text("utf-8")

        # Extract YAML frontmatter
        frontmatter_match = re.match(r"^---\s*\n(.*?)\n---\s*\n", content, re.DOTALL)
        if not frontmatter_match:
            msg = f"No YAML frontmatter found in {skill_file}"
            raise ToolError(msg)
        import yamling

        try:
            metadata = yamling.load_yaml(frontmatter_match.group(1))
        except yamling.YAMLError as e:
            msg = f"Invalid YAML frontmatter in {skill_file}: {e}"
            raise ToolError(msg) from e

        # Validate required fields
        if not isinstance(metadata, dict):
            msg = f"YAML frontmatter must be a dictionary in {skill_file}"
            raise ToolError(msg)

        name = metadata.get("name")
        description = metadata.get("description")

        if not name:
            msg = f"Missing 'name' field in {skill_file}"
            raise ToolError(msg)
        if not description:
            msg = f"Missing 'description' field in {skill_file}"
            raise ToolError(msg)

        # Validate limits
        if len(name) > SKILL_NAME_LIMIT:
            msg = f"{skill_file}: Skill name exceeds {SKILL_NAME_LIMIT} chars"
            raise ToolError(msg)
        if len(description) > SKILL_DESCRIPTION_LIMIT:
            msg = f"{skill_file}: Skill description exceeds {SKILL_DESCRIPTION_LIMIT} chars"
            raise ToolError(msg)

        return Skill(name=name, description=description, skill_path=UPath(skill_dir))

    @property
    def _error_class(self) -> type[ToolError]:
        """Error class to use for this registry."""
        return ToolError

    def _validate_item(self, item: Any) -> Skill:
        """Validate and possibly transform item before registration."""
        if not isinstance(item, Skill):
            msg = f"Expected Skill instance, got {type(item)}"
            raise ToolError(msg)
        return item

    def get_skill_instructions(self, skill_name: str) -> str:
        """Lazy load full instructions for a skill."""
        skill = self.get(skill_name)
        return skill.load_instructions()


if __name__ == "__main__":
    import asyncio
    import os

    from llmling_agent.log import configure_logging

    configure_logging()

    async def main() -> None:
        reg = SkillsRegistry()
        p = UPath(
            "github://",
            token=os.getenv("GITHUB_TOKEN"),
            username="phil65",
            org="anthropics",
            repo="skills",
        )
        print("Repository contents:")
        print([f.name for f in p.iterdir()][:5])  # Show first 5 items

        await reg.register_skills_from_path(
            p,
            token=os.getenv("GITHUB_TOKEN"),
            username="phil65",
            org="anthropics",
            repo="skills",
        )
        print(f"Found {len(reg)} skills:")
        for name, skill in reg.items():
            print(f"  - {name}: {skill.description[:60]}...")

    asyncio.run(main())



"""Models for OpenAI responses endpoint."""

from __future__ import annotations

from collections.abc import Sequence
from datetime import datetime
from typing import Any, Literal, TypedDict
from uuid import uuid4

from pydantic import Field
from schemez import Schema


class InputText(Schema):
    """Text input part."""

    type: Literal["input_text"] = "input_text"
    text: str


class InputImage(Schema):
    """Image input part."""

    type: Literal["input_image"] = "input_image"
    image_url: str


class ResponseOutputText(Schema):
    """Text output part."""

    type: Literal["output_text"] = "output_text"
    text: str
    annotations: list[dict[str, Any]] = Field(default_factory=list)


class ResponseToolCall(Schema):
    """Tool call in response."""

    type: str  # web_search_call etc
    id: str
    status: Literal["completed", "error"] = "completed"


class ResponseMessage(Schema):
    """ResponseMessage in response."""

    type: Literal["message"] = "message"
    id: str
    status: Literal["completed", "error"] = "completed"
    role: Literal["user", "assistant", "system"]
    content: list[ResponseOutputText]


class ResponseUsage(TypedDict):
    """Token usage information."""

    input_tokens: int
    input_tokens_details: dict[str, int]
    output_tokens: int
    output_tokens_details: dict[str, int]
    total_tokens: int


class ResponseRequest(Schema):
    """Request for /v1/responses endpoint."""

    model: str
    input: str | list[dict[str, Any]]
    instructions: str | None = None
    stream: bool = False
    temperature: float = 1.0
    tools: list[dict[str, Any]] = Field(default_factory=list)
    tool_choice: str = "auto"
    max_output_tokens: int | None = None
    metadata: dict[str, str] = Field(default_factory=dict)


class Response(Schema):
    """Response from /v1/responses endpoint."""

    id: str = Field(default_factory=lambda: f"resp_{uuid4().hex}")
    object: Literal["response"] = "response"
    created_at: int = Field(default_factory=lambda: int(datetime.now().timestamp()))
    status: Literal["completed", "error"] = "completed"
    error: str | None = None
    model: str
    output: Sequence[ResponseMessage | ResponseToolCall]

    # Include all the request parameters
    instructions: str | None = None
    max_output_tokens: int | None = None
    temperature: float = 1.0
    tools: list[dict[str, Any]] = Field(default_factory=list)
    tool_choice: str = "auto"
    usage: ResponseUsage | None = None
    metadata: dict[str, str] = Field(default_factory=dict)



"""MCP protocol request handlers."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

import mcp
from mcp import types

from llmling_agent.log import get_logger
from llmling_agent.mcp_server import constants


if TYPE_CHECKING:
    from pydantic import AnyUrl

    from llmling_agent_server.mcp_server.server import MCPServer


logger = get_logger(__name__)


def register_handlers(llm_server: MCPServer) -> None:  # noqa: PLR0915
    """Register all MCP protocol handlers.

    Args:
        llm_server: LLMLing server instance
    """

    @llm_server.server.set_logging_level()  # type: ignore[untyped-decorator, no-untyped-call]
    async def handle_set_level(level: mcp.LoggingLevel) -> None:
        """Handle logging level changes."""
        try:
            python_level = constants.MCP_TO_LOGGING[level]
            logger.setLevel(python_level)
            data = f"Log level set to {level}"
            await llm_server.current_session.send_log_message(level, data, logger=llm_server.name)
        except Exception as exc:
            error_data = mcp.ErrorData(
                message="Error setting log level",
                code=types.INTERNAL_ERROR,
                data=str(exc),
            )
            error = mcp.McpError(error_data)
            raise error from exc

    @llm_server.server.list_tools()  # type: ignore[untyped-decorator, no-untyped-call]
    async def handle_list_tools() -> list[types.Tool]:
        """Handle tools/list request."""
        tools = await llm_server.provider.get_tools()
        return [tool.to_mcp_tool() for tool in tools]

    @llm_server.server.call_tool()  # type: ignore[untyped-decorator]
    async def handle_call_tool(
        name: str,
        arguments: dict[str, Any] | None = None,
    ) -> list[types.TextContent]:
        """Handle tools/call request."""
        arguments = arguments or {}
        # Filter out _meta from arguments
        args = {k: v for k, v in arguments.items() if not k.startswith("_")}
        tool = await llm_server.provider.get_tool(name)
        try:
            result = await tool.execute(**args)
            return [types.TextContent(type="text", text=str(result))]
        except Exception as exc:
            logger.exception("Tool execution failed", name=name)
            error_msg = f"Tool execution failed: {exc}"
            return [types.TextContent(type="text", text=error_msg)]

    @llm_server.server.list_prompts()  # type: ignore[untyped-decorator, no-untyped-call]
    async def handle_list_prompts() -> list[types.Prompt]:
        """Handle prompts/list request."""
        return [p.to_mcp_prompt() for p in await llm_server.provider.get_prompts()]

    async def handle_get_prompt(
        name: str,
        arguments: dict[str, str] | None = None,
    ) -> types.GetPromptResult:
        """Handle prompts/get request."""
        from llmling_agent.mcp_server import conversions

        try:
            parts = await llm_server.provider.get_request_parts(name, arguments)
            messages = [msg for p in parts for msg in conversions.to_mcp_messages(p)]
            return types.GetPromptResult(messages=messages, description=name)
        except KeyError as exc:
            error_data = mcp.ErrorData(code=types.INVALID_PARAMS, message=str(exc))
            raise mcp.McpError(error_data) from exc
        except Exception as exc:
            error_data = mcp.ErrorData(code=types.INTERNAL_ERROR, message=str(exc))
            raise mcp.McpError(error_data) from exc

    @llm_server.server.progress_notification()  # type: ignore[untyped-decorator, no-untyped-call]
    async def handle_progress(
        token: str | int,
        progress: float,
        total: float | None,
        message: str | None = None,
    ) -> None:
        """Handle progress notifications from client."""
        logger.debug(
            "Progress notification",
            token=token,
            progress=progress,
            total=total,
            message=message,
        )

    @llm_server.server.subscribe_resource()  # type: ignore[untyped-decorator, no-untyped-call]
    async def handle_subscribe(uri: AnyUrl) -> None:
        """Subscribe to resource updates."""
        uri_str = str(uri)
        llm_server._subscriptions[uri_str].add(llm_server.current_session)
        logger.debug("Added subscription", uri=uri)

    @llm_server.server.unsubscribe_resource()  # type: ignore[untyped-decorator, no-untyped-call]
    async def handle_unsubscribe(uri: AnyUrl) -> None:
        """Unsubscribe from resource updates."""
        if (uri_str := str(uri)) in llm_server._subscriptions:
            llm_server._subscriptions[uri_str].discard(llm_server.current_session)
            if not llm_server._subscriptions[uri_str]:
                del llm_server._subscriptions[uri_str]
            msg = "Removed subscription"
            logger.debug(msg, uri=uri, session=llm_server.current_session)



"""Package resources for LLMling configuration."""

from __future__ import annotations

import importlib.resources
from functools import cache
from typing import Final, Literal

_RESOURCES = importlib.resources.files("llmling_agent.text_templates")

DEFAULT_SYSTEM_PROMPT: Final[str] = str(_RESOURCES / "system_prompt.jinja")
"""Path to the default system prompt template."""

FormatStyle = Literal["simple", "detailed", "markdown"]


@cache
def get_system_prompt() -> str:
    """Get the default system prompt template content."""
    return (_RESOURCES / "system_prompt.jinja").read_text()


@cache
def get_tool_call_template(style: FormatStyle) -> str:
    """Get tool call template by style."""
    files = {
        "simple": "tool_call_simple.jinja",
        "detailed": "tool_call_default.jinja",
        "markdown": "tool_call_markdown.jinja",
    }
    return (_RESOURCES / files[style]).read_text()


__all__ = [
    "DEFAULT_SYSTEM_PROMPT",
    "FormatStyle",
    "get_system_prompt",
    "get_tool_call_template",
]



"""Fetch git diff for a specific commit."""

from __future__ import annotations

import uuid

from pydantic_ai import UserPromptPart
from slashed import CommandContext  # noqa: TC002

from llmling_agent.log import get_logger
from llmling_agent.messaging.context import NodeContext  # noqa: TC001
from llmling_agent_commands.base import NodeCommand
from llmling_agent_server.acp_server.session import ACPSession  # noqa: TC001


logger = get_logger(__name__)


class GitDiffCommand(NodeCommand):
    """Fetch git diff for a specific commit.

    Executes git diff client-side to get the complete diff
    and displays it in a structured format.
    """

    name = "git-diff"
    category = "docs"

    async def execute_command(
        self,
        ctx: CommandContext[NodeContext[ACPSession]],
        commit: str,
        *,
        base_commit: str | None = None,
        cwd: str | None = None,
    ) -> None:
        """Fetch git diff for a specific commit.

        Args:
            ctx: Command context with ACP session
            commit: Commit hash or reference to diff
            base_commit: Base commit to compare against (defaults to commit~1)
            cwd: Working directory to run git in (defaults to session cwd)
        """
        session = ctx.context.data
        assert session

        # Generate tool call ID
        tool_call_id = f"git-diff-{uuid.uuid4().hex[:8]}"

        try:
            # Check if we have terminal access for running git
            if not (
                session.client_capabilities
                and session.client_capabilities.terminal
                and session.acp_agent.terminal_access
            ):
                await session.notifications.send_agent_text(
                    "❌ **Terminal access not available for git operations**"
                )
                return

            # Build git diff command
            if base_commit:
                git_command = ["git", "diff", base_commit, commit]
                display_title = f"Git diff: {base_commit}..{commit}"
            else:
                git_command = ["git", "diff", f"{commit}~1", commit]
                display_title = f"Git diff: {commit}"

            # Start tool call
            await session.notifications.tool_call_start(
                tool_call_id=tool_call_id,
                title=display_title,
                kind="execute",
            )

            # Run git diff command
            output, exit_code = await session.requests.run_command(
                command=git_command[0],
                args=git_command[1:],
                cwd=cwd or session.cwd,
            )

            # Check if git command succeeded
            if exit_code != 0:
                await session.notifications.tool_call_progress(
                    tool_call_id=tool_call_id,
                    status="failed",
                    title=f"Git diff failed (exit code: {exit_code})",
                )
                return

            diff_content = output.strip()
            if not diff_content:
                await session.notifications.tool_call_progress(
                    tool_call_id=tool_call_id,
                    status="completed",
                    title="No changes found in diff",
                )
                return

            # Stage the diff content for use in agent context
            staged_part = UserPromptPart(content=f"Git diff for {display_title}:\n\n{diff_content}")
            session.add_staged_parts([staged_part])

            # Send successful result - wrap in code block for proper display
            staged_count = session.get_staged_parts_count()
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="completed",
                title=f"Git diff fetched and staged ({staged_count} total parts)",
                content=[f"```diff\n{diff_content}\n```"],
            )

        except Exception as e:
            logger.exception("Unexpected error fetching git diff", commit=commit)
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="failed",
                title=f"Error: {e}",
            )



"""Function discovery and metadata handling."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import dataclass, field
import inspect
from typing import TYPE_CHECKING, Any

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Sequence

    from llmling_agent.common_types import AnyCallable

logger = get_logger(__name__)


@dataclass
class NodeFunction:
    """Metadata for a function that uses nodes."""

    func: AnyCallable
    """The actual function to execute."""

    depends_on: list[str] = field(default_factory=list)
    """Names of functions this one depends on."""

    deps: Any = None
    """Node dependencies."""

    default_inputs: dict[str, Any] = field(default_factory=dict)
    """Default parameter values."""

    name: str = field(init=False)
    """Function name (from function.__name__)."""

    def __post_init__(self) -> None:
        """Set name and validate dependencies."""
        self.name = self.func.__name__
        # Extract default inputs from function signature

        sig = inspect.signature(self.func)
        self.default_inputs = {
            name: param.default
            for name, param in sig.parameters.items()
            if param.default is not param.empty
        }
        logger.debug("Registered node function", name=self.name, deps=self.depends_on)


def node_function(
    func: Callable[..., Any] | None = None,
    *,
    deps: Any | None = None,
    depends_on: str | Sequence[str | Callable[..., Any]] | Callable[..., Any] | None = None,
) -> Callable[..., Any]:
    """Mark a function for automatic node execution.

    Can be used as simple decorator or with arguments:

    @node_function
    async def func(): ...

    @node_function(order=1, depends_on="other_func")
    async def func(): ...

    Args:
        func: Function to mark
        deps: Dependencies to inject into all Agent parameters
        depends_on: Names of functions this one depends on

    Returns:
        Decorated function
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        match depends_on:
            case None:
                depends_on_ = []
            case str():
                depends_on_ = [depends_on]
            case Callable():
                depends_on_ = [depends_on.__name__]

            case [*items]:
                depends_on_ = [
                    i.__name__ if isinstance(i, Callable) else str(i)  # type: ignore[union-attr, arg-type]
                    for i in items
                ]
            case _:
                msg = f"Invalid depends_on: {depends_on}"
                raise ValueError(msg)
        # TODO: we still need to inject the deps in execution part.
        metadata = NodeFunction(func=func, depends_on=depends_on_ or [], deps=deps)
        func._node_function = metadata  # type: ignore
        return func

    return decorator(func) if func is not None else decorator



"""A2A server implementation following BaseServer pattern."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Sequence
from contextlib import asynccontextmanager
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Self, TypeVar
import uuid

from fasta2a.schema import (  # type: ignore[import-untyped]
    AgentCapabilities,
    AgentCard,
    a2a_request_ta,
    a2a_response_ta,
    agent_card_ta,
)
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.responses import FileResponse, Response
from starlette.routing import Route
import uvicorn

from llmling_agent.log import get_logger
from llmling_agent.messaging import MessageHistory
from llmling_agent.messaging.messages import ChatMessage
from llmling_agent_server import BaseServer
from llmling_agent_server.a2a_server.storage import SimpleStorage


if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Sequence

    from fasta2a.schema import A2AResponse, AgentProvider, Skill
    from starlette.middleware import Middleware
    from starlette.requests import Request
    from starlette.routing import Route
    from starlette.types import ExceptionHandler, Receive, Scope, Send

    from llmling_agent import Agent, AgentPool
    from llmling_agent_server.a2a_server.a2a_types import A2ARequest


logger = get_logger(__name__)

# AgentWorker output type needs to be invariant for use
# in both parameter and return positions
WorkerOutputT = TypeVar("WorkerOutputT")

AgentDepsT = Any
OutputDataT = Any


class FastA2A(Starlette):
    """The main class for the FastA2A library."""

    def __init__(
        self,
        *,
        storage: SimpleStorage,
        broker: SimpleBroker,
        # Agent card
        name: str | None = None,
        url: str = "http://localhost:8000",
        version: str = "1.0.0",
        description: str | None = None,
        provider: AgentProvider | None = None,
        skills: list[Skill] | None = None,
        # Starlette
        debug: bool = False,
        routes: Sequence[Route] | None = None,
        middleware: Sequence[Middleware] | None = None,
        exception_handlers: dict[Any, ExceptionHandler] | None = None,
        lifespan: Any = None,
    ) -> None:
        if lifespan is None:
            lifespan = _default_lifespan

        super().__init__(
            debug=debug,
            routes=routes,
            middleware=middleware,
            exception_handlers=exception_handlers,
            lifespan=lifespan,
        )

        self.name = name or "My Agent"
        self.url = url
        self.version = version
        self.description = description
        self.provider = provider
        self.skills = skills or []
        # NOTE: For now, I don't think there's any reason to support any other
        # input/output modes.
        self.default_input_modes = ["application/json"]
        self.default_output_modes = ["application/json"]

        self.task_manager = TaskManager(broker=broker, storage=storage)

        # Setup
        self._agent_card_json_schema: bytes | None = None
        self.router.add_route(
            "/.well-known/agent-card.json",
            self._agent_card_endpoint,
            methods=["HEAD", "GET", "OPTIONS"],
        )
        self.router.add_route("/", self._agent_run_endpoint, methods=["POST"])
        self.router.add_route("/docs", self._docs_endpoint, methods=["GET"])

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "http" and not self.task_manager.is_running:
            raise RuntimeError("TaskManager was not properly initialized.")
        await super().__call__(scope, receive, send)

    async def _agent_card_endpoint(self, request: Request) -> Response:
        if self._agent_card_json_schema is None:
            agent_card = AgentCard(
                name=self.name,
                description=self.description or "An AI agent exposed as an A2A agent.",
                url=self.url,
                version=self.version,
                protocol_version="0.3.0",
                skills=self.skills,
                default_input_modes=self.default_input_modes,
                default_output_modes=self.default_output_modes,
                capabilities=AgentCapabilities(
                    streaming=False,
                    push_notifications=False,
                    state_transition_history=False,
                ),
            )
            if self.provider is not None:
                agent_card["provider"] = self.provider
            self._agent_card_json_schema = agent_card_ta.dump_json(agent_card, by_alias=True)
        return Response(content=self._agent_card_json_schema, media_type="application/json")

    async def _docs_endpoint(self, request: Request) -> Response:
        """Serve the documentation interface."""
        docs_path = Path(__file__).parent / "static" / "docs.html"
        return FileResponse(docs_path, media_type="text/html")

    async def _agent_run_endpoint(self, request: Request) -> Response:
        """Main endpoint for the A2A server.

        Although the specification allows freedom of choice and implementation,
        I'm pretty sure about some decisions.

        1. The server will always either send a "submitted" or a "failed" on
           `message/send`. Never a "completed" on the first message.
        2. There are three possible ends for the task:
            2.1. The task was "completed" successfully.
            2.2. The task was "canceled".
            2.3. The task "failed".
        3. The server will send a "working" on the first chunk on
           `tasks/pushNotification/get`.
        """
        data = await request.body()
        a2a_request = a2a_request_ta.validate_json(data)

        if a2a_request["method"] == "message/send":
            jsonrpc_response = await self.task_manager.send_message(a2a_request)  # pyright: ignore[reportArgumentType]
        elif a2a_request["method"] == "tasks/get":
            jsonrpc_response = await self.task_manager.get_task(a2a_request)  # pyright: ignore[reportArgumentType]
        elif a2a_request["method"] == "tasks/cancel":
            jsonrpc_response = await self.task_manager.cancel_task(a2a_request)  # pyright: ignore[reportArgumentType]
        else:
            raise NotImplementedError(f"Method {a2a_request['method']} not implemented.")
        return Response(
            content=a2a_response_ta.dump_json(jsonrpc_response, by_alias=True),
            media_type="application/json",
        )


class TaskManager:
    """Task manager for A2A server."""

    def __init__(self, *, broker: SimpleBroker, storage: SimpleStorage) -> None:
        self.broker = broker
        self.storage = storage
        self.is_running = False

    async def __aenter__(self) -> Self:
        self.is_running = True
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        self.is_running = False

    async def send_message(self, request: A2ARequest) -> A2AResponse:
        """Handle message/send requests."""
        task_id = str(uuid.uuid4())
        # Store task in storage for processing
        await self.storage.store_task(task_id, request)
        await self.broker.submit_task(task_id)
        return {  # pyright: ignore[reportReturnType]
            "jsonrpc": "2.0",
            "id": request.get("id"),
            "result": {"status": "submitted", "task_id": task_id},
        }

    async def get_task(self, request: A2ARequest) -> A2AResponse:
        """Handle tasks/get requests."""
        task_id = request.get("params", {}).get("task_id")
        if not task_id or not isinstance(task_id, str):
            return {  # pyright: ignore[reportReturnType]
                "jsonrpc": "2.0",
                "id": request.get("id"),
                "result": {
                    "status": "invalid_request",
                    "result": None,
                    "error": "Missing task_id",
                },
            }
        task_status = await self.storage.get_task_status(task_id)
        return {  # pyright: ignore[reportReturnType]
            "jsonrpc": "2.0",
            "id": request.get("id"),
            "result": {
                "status": task_status["status"],
                "result": task_status["result"],
                "error": task_status["error"],
            },
        }

    async def cancel_task(self, request: A2ARequest) -> A2AResponse:
        """Handle tasks/cancel requests."""
        task_id = request.get("params", {}).get("task_id")
        if not task_id or not isinstance(task_id, str):
            return {  # pyright: ignore[reportReturnType]
                "jsonrpc": "2.0",
                "id": request.get("id"),
                "result": {"status": "invalid_request"},
            }
        await self.storage.cancel_task(task_id)
        return {  # pyright: ignore[reportReturnType]
            "jsonrpc": "2.0",
            "id": request.get("id"),
            "result": {"status": "cancelled"},
        }


class SimpleBroker:
    """Simple in-memory broker for A2A tasks."""

    def __init__(self, worker: AgentWorker | None = None) -> None:
        self.worker = worker
        self.tasks: list[str] = []

    async def submit_task(self, task_id: str) -> None:
        """Submit a task for processing."""
        self.tasks.append(task_id)
        if self.worker:
            # Process immediately for simplicity
            await self.worker.run_task(task_id)

    def set_worker(self, worker: AgentWorker) -> None:
        """Set the worker for this broker."""
        self.worker = worker


@asynccontextmanager
async def _default_lifespan(app: FastA2A) -> AsyncIterator[None]:
    async with app.task_manager:
        yield


@asynccontextmanager
async def worker_lifespan(
    app: FastA2A, worker: AgentWorker, agent: Agent[AgentDepsT, OutputDataT]
) -> AsyncIterator[None]:
    """Custom lifespan that runs the worker during application startup."""
    async with app.task_manager, agent, await worker.run():
        yield


@dataclass
class AgentWorker:
    """A worker that uses an agent to execute tasks."""

    agent: Agent[AgentDepsT, OutputDataT]
    storage: SimpleStorage
    broker: SimpleBroker | None = None

    async def run(self) -> AgentWorker:
        """Run the worker."""
        return self

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: object,
    ) -> None:
        pass

    async def run_task(self, task_id: str) -> None:
        """Run a task using the agent."""
        task_data = await self.storage.get_task(task_id)
        if not task_data:
            raise ValueError(f"Task {task_id} not found")

        context_id = task_data["context_id"]
        await self.storage.update_task_status(task_id, "working")

        try:
            # Load conversation history as ChatMessages
            if context_id:
                message_history = await self.storage.load_context(context_id)
            else:
                message_history = []
            # Convert A2A message to ChatMessage and add to history
            new_message = self._a2a_to_chat_message(task_data["data"])
            # Run the agent with the new message as prompt and previous history as context
            history = MessageHistory(messages=message_history)
            result = await self.agent.run(new_message, message_history=history)

            message_history.append(new_message)
            message_history.append(result)
            if context_id:
                await self.storage.update_context(context_id, message_history)

            # Convert result to A2A format
            a2a_result = self._chat_message_to_a2a(result)

            await self.storage.update_task_status(task_id, "completed", result=a2a_result)
        except Exception as e:  # noqa: BLE001
            await self.storage.update_task_status(task_id, "failed", error=str(e))

    async def cancel_task(self, task_id: str) -> None:
        """Cancel a task."""
        await self.storage.update_task_status(task_id, "cancelled")

    def _a2a_to_chat_message(self, task_data: A2ARequest) -> ChatMessage[Any]:
        """Convert A2A task data to ChatMessage."""
        # Extract message content from A2A request
        messages = task_data.get("params", {}).get("messages", [])
        content = ""
        if messages:
            parts = messages[0].get("parts", [])
            for part in parts:
                if part.get("kind") == "text":
                    content = part.get("text", "")
                    break

        return ChatMessage.user_prompt(
            message=content,
            conversation_id=task_data.get("params", {}).get("context_id"),
        )

    def _chat_message_to_a2a(self, message: ChatMessage[Any]) -> dict[str, Any]:
        """Convert ChatMessage to A2A format."""
        return {
            "messages": [
                {
                    "role": "agent",
                    "parts": [
                        {
                            "kind": "text",
                            "text": str(message.content),
                        }
                    ],
                }
            ]
        }


class A2AServer(BaseServer):
    """A2A server wrapper following BaseServer pattern.

    This server wraps the FastA2A Starlette application and adapts it
    to follow our standard BaseServer interface for consistent lifecycle management.
    """

    def __init__(
        self,
        pool: AgentPool,
        *,
        name: str | None = None,
        host: str = "0.0.0.0",
        port: int = 8000,
        # A2A specific options
        agent_name: str | None = None,
        url: str | None = None,
        version: str = "1.0.0",
        description: str | None = None,
        # Server options
        raise_exceptions: bool = False,
        **kwargs: Any,
    ):
        """Initialize A2A server.

        Args:
            pool: AgentPool containing available agents
            name: Optional Server name (auto-generated if None)
            host: Host to bind server to
            port: Port to bind server to
            agent_name: Name for the A2A agent card
            url: URL for the A2A agent card
            version: Version for the A2A agent card
            description: Description for the A2A agent card
            raise_exceptions: Whether to raise exceptions during server start
            **kwargs: Additional arguments passed to FastA2A
        """
        super().__init__(pool, name=name, raise_exceptions=raise_exceptions)
        self.host = host
        self.port = port
        self.agent_name = agent_name or "LLMling Agent"
        self.url = url or f"http://{host}:{port}"
        self.version = version
        self.description = description
        self.kwargs = kwargs

        # Get the first agent from the pool for now
        # TODO: Support multiple agents or agent selection
        if not self.pool.agents:
            raise ValueError("No agents available in pool")

        agent = next(iter(self.pool.agents.values()))

        # Create A2A components
        storage = SimpleStorage()
        worker = AgentWorker(agent=agent, storage=storage)
        broker = SimpleBroker(worker=worker)
        worker.broker = broker

        lifespan = partial(worker_lifespan, worker=worker, agent=agent)

        # Create the FastA2A app
        self._app = FastA2A(
            storage=storage,
            broker=broker,
            name=self.agent_name,
            url=self.url,
            version=self.version,
            description=self.description,
            lifespan=lifespan,
            **self.kwargs,
        )

    async def _start_async(self) -> None:
        """Start the A2A server (blocking async - runs until stopped)."""
        config = uvicorn.Config(
            self._app,
            host=self.host,
            port=self.port,
            log_level="info",
            ws="websockets-sansio",
        )
        server = uvicorn.Server(config)
        await server.serve()

    @property
    def app(self) -> FastA2A:
        """Get the underlying FastA2A application."""
        return self._app


if __name__ == "__main__":
    import httpx

    from llmling_agent import AgentPool

    async def test_client() -> None:
        """Test the A2A server with a basic request."""
        async with httpx.AsyncClient() as client:
            # Test agent card endpoint
            response = await client.get("http://localhost:8000/.well-known/agent-card.json")
            print("Agent card:", response.json())

            # Test docs endpoint
            response = await client.get("http://localhost:8000/docs")
            print("Docs status:", response.status_code)

    async def main() -> None:
        """Run server and test client."""
        pool = AgentPool()
        await pool.add_agent("test-agent", model="openai:gpt-5-nano")

        async with (
            A2AServer(
                pool,
                host="0.0.0.0",
                port=8000,
                agent_name="Test LLMling Agent",
                description="A test agent for A2A protocol",
            ) as server,
            server.run_context(),
        ):
            await asyncio.sleep(1)  # Wait for server to start
            await test_client()

    asyncio.run(main())



"""Utils for documentation."""

from __future__ import annotations

import ast
from pathlib import Path
from typing import Literal

import mknodes as mk


DocStyle = Literal["simple", "full"]


def create_example_doc(name: str, *, style: DocStyle = "full") -> mk.MkContainer:
    """Create documentation for an example file.

    Args:
        name: Name of example file (e.g. "download_agents.py")
        examples_dir: Directory containing examples (default: examples/)
        style: Documentation style:
            - simple: Just the code
            - full: Title, description from docstring, code, etc.

    Returns:
        Container with all documentation elements
    """
    path = Path("src/llmling_agent_docs/examples") / name
    if not path.exists():
        raise FileNotFoundError(f"Example {name} not found")

    container = mk.MkContainer()

    if style == "full":
        # Extract title/description from example's docstring
        title = path.stem.replace("_", " ").title()
        container += mk.MkHeader(title, level=2)

        # Add description from docstring if available
        if docstring := ast.get_docstring(ast.parse(path.read_text(encoding="utf-8"))):
            container += docstring

    # Add the code itself
    container += mk.MkCode(path.read_text(encoding="utf-8"), language="python")
    return container



"""VFS registry filesystem toolset implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING

from llmling_agent.resource_providers import StaticResourceProvider


if TYPE_CHECKING:
    from llmling_agent.agent.context import AgentContext


async def vfs_list(  # noqa: D417
    ctx: AgentContext,
    path: str = "",
    *,
    pattern: str = "**/*",
    recursive: bool = True,
    include_dirs: bool = False,
    exclude: list[str] | None = None,
    max_depth: int | None = None,
) -> str:
    """List contents of a resource path.

    Args:
        path: Resource path to query (e.g., "docs", "docs://guides", "data://files")
        pattern: Glob pattern to match files against
        recursive: Whether to search subdirectories
        include_dirs: Whether to include directories in results
        exclude: List of patterns to exclude
        max_depth: Maximum directory depth for recursive search

    Returns:
        Formatted list of matching files and directories
    """
    registry = ctx.definition.vfs_registry

    # If no path given, list all resources
    if not path:
        resources = list(registry.keys())
        return "Available resources:\n" + "\n".join(f"- {name}" for name in resources)

    try:
        files = await registry.query(
            path,
            pattern=pattern,
            recursive=recursive,
            include_dirs=include_dirs,
            exclude=exclude,
            max_depth=max_depth,
        )
        return "\n".join(files) if files else f"No files found in {path}"
    except (OSError, ValueError, KeyError) as e:
        return f"Error listing {path}: {e}"


async def vfs_read(  # noqa: D417
    ctx: AgentContext,
    path: str,
    *,
    encoding: str = "utf-8",
    recursive: bool = True,
    exclude: list[str] | None = None,
    max_depth: int | None = None,
) -> str:
    """Read content from a resource path.

    Args:
        path: Resource path to read (e.g., "docs://file.md", "data://folder")
        encoding: Text encoding for binary content
        recursive: For directories, whether to read recursively
        exclude: For directories, patterns to exclude
        max_depth: For directories, maximum depth to read

    Returns:
        File content or concatenated directory contents
    """
    registry = ctx.definition.vfs_registry

    try:
        return await registry.get_content(
            path,
            encoding=encoding,
            recursive=recursive,
            exclude=exclude,
            max_depth=max_depth,
        )
    except (OSError, ValueError, KeyError) as e:
        return f"Error reading {path}: {e}"


async def vfs_info(ctx: AgentContext) -> str:
    """Get information about all configured resources.

    Returns:
        Formatted information about available resources
    """
    registry = ctx.definition.vfs_registry

    if registry.is_empty:
        return "No resources configured"

    sections = ["## Configured Resources\n"]

    for name in registry:
        try:
            fs = registry[name]
            fs_type = fs.__class__.__name__
            sections.append(f"- **{name}**: {fs_type}")
        except (OSError, AttributeError) as e:
            sections.append(f"- **{name}**: Error - {e}")

    # Add union filesystem info
    try:
        union_fs = registry.get_fs()
        sections.append(f"\n**Union filesystem**: {union_fs.__class__.__name__}")
        sections.append(f"**Total resources**: {len(registry)}")
    except (OSError, AttributeError) as e:
        sections.append(f"\n**Union filesystem error**: {e}")

    return "\n".join(sections)


class VFSTools(StaticResourceProvider):
    """Provider for VFS registry filesystem tools."""

    def __init__(self, name: str = "vfs") -> None:
        tools = [
            self.create_tool(vfs_list, category="search", read_only=True, idempotent=True),
            self.create_tool(vfs_read, category="read", read_only=True, idempotent=True),
            self.create_tool(vfs_info, category="read", read_only=True, idempotent=True),
        ]
        super().__init__(name=name, tools=tools)



"""Signature utils."""

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, get_origin

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


logger = get_logger(__name__)


def create_modified_signature(
    fn_or_sig: Callable[..., Any] | inspect.Signature,
    *,
    remove: str | list[str] | None = None,
    inject: dict[str, type] | None = None,
) -> inspect.Signature:
    """Create a modified signature by removing specified parameters / injecting new ones.

    Args:
        fn_or_sig: The function or signature to modify.
        remove: The parameter(s) to remove.
        inject: The parameter(s) to inject.

    Returns:
        The modified signature.
    """
    sig = fn_or_sig if isinstance(fn_or_sig, inspect.Signature) else inspect.signature(fn_or_sig)
    rem_keys = [remove] if isinstance(remove, str) else remove or []
    new_params = [p for p in sig.parameters.values() if p.name not in rem_keys]
    if inject:
        injected_params = []
        for k, v in inject.items():
            injected_params.append(
                inspect.Parameter(k, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=v)
            )
        new_params = injected_params + new_params
    return sig.replace(parameters=new_params)


def modify_signature(
    fn: Callable[..., Any],
    *,
    remove: str | list[str] | None = None,
    inject: dict[str, type] | None = None,
) -> None:
    new_sig = create_modified_signature(fn, remove=remove, inject=inject)
    update_signature(fn, new_sig)


def update_signature(fn: Callable[..., Any], signature: inspect.Signature) -> None:
    fn.__signature__ = signature  # type: ignore
    fn.__annotations__ = {
        name: param.annotation for name, param in signature.parameters.items()
    } | {"return": signature.return_annotation}


def create_bound_callable(  # noqa: PLR0915
    original_callable: Callable[..., Any],
    by_name: dict[str, Any] | None = None,
    by_type: dict[type, Any] | None = None,
    bind_kwargs: bool = False,
) -> Callable[..., Awaitable[Any]]:
    """Create a wrapper that pre-binds parameters by name or type.

    Parameters are bound by their position in the function signature. Only
    positional and positional-or-keyword parameters can be bound by default.
    If bind_kwargs=True, keyword-only parameters can also be bound using the
    same by_name/by_type logic. Binding by name takes priority over binding by type.

    Args:
        original_callable: The original callable that may need parameter binding
        by_name: Parameters to bind by exact parameter name
        by_type: Parameters to bind by parameter type annotation
        bind_kwargs: Whether to also bind keyword-only parameters

    Returns:
        New callable with parameters pre-bound and proper introspection

    Raises:
        ValueError: If the callable's signature cannot be inspected
    """
    try:
        sig = inspect.signature(original_callable)
    except (ValueError, TypeError) as e:
        msg = f"Cannot inspect signature of {original_callable}. Ensure callable is inspectable."
        raise ValueError(msg) from e

    # Build position-to-value mapping for positional binding
    context_values = {}
    # Build name-to-value mapping for keyword-only binding
    kwarg_bindings = {}

    for i, param in enumerate(sig.parameters.values()):
        # Bind positional and positional-or-keyword parameters
        if param.kind in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
        ):
            # Bind by name first (higher priority)
            if by_name and param.name in by_name:
                context_values[i] = by_name[param.name]
            # Then bind by type if not already bound
            elif by_type and _find_matching_type(param.annotation, by_type) is not None:
                context_values[i] = _find_matching_type(param.annotation, by_type)
        # Bind keyword-only parameters if enabled
        elif bind_kwargs and param.kind == inspect.Parameter.KEYWORD_ONLY:
            # Bind by name first (higher priority)
            if by_name and param.name in by_name:
                kwarg_bindings[param.name] = by_name[param.name]
            # Then bind by type if not already bound
            elif by_type and _find_matching_type(param.annotation, by_type) is not None:
                kwarg_bindings[param.name] = _find_matching_type(param.annotation, by_type)

    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        # Filter out kwargs that would conflict with bound parameters
        param_names = list(sig.parameters.keys())
        bound_param_names = {param_names[i] for i in context_values}
        bound_kwarg_names = set(kwarg_bindings.keys())
        filtered_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in bound_param_names and k not in bound_kwarg_names
        }

        # Add bound keyword-only parameters
        filtered_kwargs.update(kwarg_bindings)

        # Build new_args with context values at correct positions
        new_args = []
        arg_index = 0
        for param_index in range(len(sig.parameters)):
            if param_index in context_values:
                new_args.append(context_values[param_index])
            elif arg_index < len(args):
                new_args.append(args[arg_index])
                arg_index += 1

        # Add any remaining positional args
        if arg_index < len(args):
            new_args.extend(args[arg_index:])

        if inspect.iscoroutinefunction(original_callable):
            return await original_callable(*new_args, **filtered_kwargs)
        return original_callable(*new_args, **filtered_kwargs)

    # Preserve introspection attributes
    wrapper.__name__ = getattr(original_callable, "__name__", "wrapper")
    wrapper.__doc__ = getattr(original_callable, "__doc__", None)
    wrapper.__module__ = getattr(original_callable, "__module__", None)  # type: ignore[assignment]
    wrapper.__wrapped__ = original_callable  # type: ignore[attr-defined]
    wrapper.__llmling_wrapped__ = original_callable  # type: ignore[attr-defined]

    # Create modified signature without context parameters
    try:
        params = list(sig.parameters.values())
        # Remove parameters at context positions and bound kwargs
        context_positions = set(context_values.keys())
        bound_kwarg_names = set(kwarg_bindings.keys())
        new_params = [
            param
            for i, param in enumerate(params)
            if i not in context_positions and param.name not in bound_kwarg_names
        ]
        new_sig = sig.replace(parameters=new_params)
        wrapper.__signature__ = new_sig  # type: ignore[attr-defined]
        wrapper.__annotations__ = {
            name: param.annotation for name, param in new_sig.parameters.items()
        }
        if sig.return_annotation != inspect.Signature.empty:
            wrapper.__annotations__["return"] = sig.return_annotation

    except (ValueError, TypeError):
        logger.debug("Failed to update wrapper signature", original=original_callable)

    return wrapper


def _find_matching_type(param_annotation: Any, by_type: dict[type, Any]) -> Any | None:
    """Find a matching type binding for the given parameter annotation.

    Supports exact matching and generic origin matching.

    Args:
        param_annotation: The parameter's type annotation
        by_type: Dictionary of type bindings

    Returns:
        The bound value if a match is found, None otherwise
    """
    # First try exact match
    if param_annotation in by_type:
        return by_type[param_annotation]

    # Then try origin type matching for generics
    param_origin = get_origin(param_annotation)
    if param_origin is not None and param_origin in by_type:
        return by_type[param_origin]

    return None



"""ACP-specific slash commands for session management."""

from __future__ import annotations

from slashed import CommandContext  # noqa: TC002

from llmling_agent.messaging.context import NodeContext  # noqa: TC001
from llmling_agent_commands.base import NodeCommand
from llmling_agent_server.acp_server.session import ACPSession  # noqa: TC001


class ListSessionsCommand(NodeCommand):
    """List all available ACP sessions.

    Shows:
    - Session ID and status (active/stored)
    - Agent name and working directory
    - Creation time and message count
    - Storage information
    """

    name = "list-sessions"
    category = "acp"

    async def execute_command(
        self,
        ctx: CommandContext[NodeContext[ACPSession]],
        *,
        active: bool = False,
        stored: bool = False,
    ) -> None:
        """List available ACP sessions.

        Args:
            ctx: Command context with ACP session
            active: Show only active sessions
            stored: Show only stored sessions
        """
        session = ctx.context.data
        assert session
        # Check if we have access to session manager
        if not session.manager:
            await ctx.print("❌ **Session manager not available**")
            return

        # If no filter specified, show both
        if not active and not stored:
            active = stored = True

        try:
            output_lines = ["## 📋 ACP Sessions\n"]

            # Show active sessions
            if active:
                output_lines.append("### 🟢 Active Sessions")
                active_sessions = session.manager._sessions

                if not active_sessions:
                    output_lines.append("*No active sessions*\n")
                else:
                    for session_id, sess in active_sessions.items():
                        agent_name = getattr(sess, "current_agent_name", "unknown")
                        cwd = getattr(sess, "cwd", "unknown")
                        msg_count = len(getattr(sess, "_conversation_history", []))

                        output_lines.append(f"- **{session_id}**")
                        output_lines.append(f"  - Agent: `{agent_name}`")
                        output_lines.append(f"  - Directory: `{cwd}`")
                        output_lines.append(f"  - Messages: {msg_count}")
                    output_lines.append("")

            # # Show stored sessions
            # if stored and session.manager._persistent_manager:
            #     output_lines.append("### 💾 Stored Sessions")

            #     try:
            #         stored_sessions = (
            #             await session.manager._persistent_manager.store.list_sessions()
            #         )

            #         if not stored_sessions:
            #             output_lines.append("*No stored sessions*\n")
            #         else:
            #             for session_id in stored_sessions:
            #                 store = session.manager._persistent_manager.store
            #                 session_data = await store.load_session(session_id)
            #                 if session_data:
            #                     msg_count = len(session_data.conversation)
            #                     created = session_data.metadata.get(
            #                         "created_at", "unknown"
            #                     )

            #                     output_lines.append(f"- **{session_id}**")
            #                     output_lines.append(
            #                         f"  - Agent: `{session_data.agent_name or 'unknown'}`"  # noqa: E501
            #                     )
            #                     output_lines.append(
            #                         f"  - Directory: `{session_data.cwd}`"
            #                     )
            #                     output_lines.append(f"  - Messages: {msg_count}")
            #                     output_lines.append(f"  - Created: {created}")
            #             output_lines.append("")
            #     except Exception as e:
            #         output_lines.append(f"*Error loading stored sessions: {e}*\n")

            await ctx.print("\n".join(output_lines))

        except Exception as e:  # noqa: BLE001
            await ctx.print(f"❌ **Error listing sessions:** {e}")


# class LoadSessionCommand(SlashedCommand):
#     """Load a previous ACP session with conversation replay.

#     This command will:
#     1. Look up the session by ID
#     2. Replay the entire conversation history
#     3. Restore the session context (agent, working directory, MCP servers)
#     """

#     name = "load-session"
#     category = "acp"

#     async def execute_command(
#         self,
#         ctx: CommandContext[ACPCommandContext],
#         session_id: str,
#         *,
#         preview: bool = False,
#         no_replay: bool = False,
#     ):
#         """Load a previous ACP session.

#         Args:
#             ctx: Command context with ACP session
#             session_id: Session identifier to load
#             preview: Show session info without loading
#             no_replay: Load session without replaying conversation
#         """
#         session = ctx.context.session
#         if not session.manager:
#             await ctx.print("❌ **Session manager not available**")
#             return

#         try:
#             # Check if session exists
#             if session.manager._persistent_manager:
#                 session_data = (
#                     await session.manager._persistent_manager.load_session_data(
#                         session_id
#                     )
#                 )
#                 if not session_data:
#                     await ctx.print(f"❌ **Session not found:** `{session_id}`")
#                     return

#                 if preview:
#                     # Show session preview
#                     msg_count = len(session_data.conversation)
#                     created = session_data.metadata.get("created_at", "unknown")
#                     mcp_count = len(session_data.mcp_servers)

#                     preview_lines = [
#                         f"## 📋 Session Preview: `{session_id}`\n",
#                         f"**Agent:** `{session_data.agent_name or 'unknown'}`",
#                         f"**Directory:** `{session_data.cwd}`",
#                         f"**Messages:** {msg_count}",
#                         f"**MCP Servers:** {mcp_count}",
#                         f"**Created:** {created}",
#                     ]

#                     if session_data.metadata:
#                         metadata_json = json.dumps(session_data.metadata, indent=2)
#                         preview_lines.append(
#                             f"**Metadata:** ```json\n{metadata_json}\n```"
#                         )

#                     await ctx.print("\n".join(preview_lines))
#                     return

#                 # Actually load the session
#                 if no_replay:
#                     await ctx.print(
#                         f"🔄 **Loading session `{session_id}` without replay...**"
#                     )
#                     await ctx.print(
#                         f"✅ **Session `{session_id}` is available for loading**"
#                     )
#                 else:
#                     load_msg = f"🔄 **Loading session `{session_id}` with replay...**"
#                     await ctx.print(load_msg)

#                     msg_count = len(session_data.conversation)
#                     await ctx.print(f"📽️ **Replaying {msg_count} messages...**")
#                     await ctx.print(
#                         f"✅ **Session `{session_id}` loaded successfully**"
#                     )
#             else:
#                 await ctx.print("❌ **Session persistence not enabled**")

#         except Exception as e:
#             await ctx.print(f"❌ **Error loading session:** {e}")


# class SaveSessionCommand(SlashedCommand):
#     """Save the current ACP session to persistent storage.

#     This will save:
#     - Complete conversation history
#     - Current agent configuration
#     - Working directory and MCP server setup
#     - Session metadata

#     The session can later be loaded with /load-session.
#     """

#     name = "save-session"
#     category = "acp"

#     async def execute_command(
#         self,
#         ctx: CommandContext[ACPCommandContext],
#         *,
#         description: str | None = None,
#     ):
#         """Save the current ACP session.

#         Args:
#             ctx: Command context with ACP session
#             description: Optional description for the session
#         """
#         session = ctx.context.session

#         if not session.manager:
#             await ctx.print("❌ **Session manager not available**")
#             return

#         try:
#             if session.manager._persistent_manager:
#                 await session.manager._persistent_manager.save_session(session)

#                 msg_count = len(getattr(session, "_conversation_history", []))
#                 await ctx.print(
#                     f"💾 **Session `{session.session_id}` saved successfully**"
#                 )
#                 await ctx.print(f"📊 **Saved {msg_count} messages**")

#                 if description:
#                     await ctx.print(f"📝 **Description:** {description}")
#             else:
#                 await ctx.print("❌ **Session persistence not enabled**")

#         except Exception as e:
#             await ctx.print(f"❌ **Error saving session:** {e}")


# class DeleteSessionCommand(SlashedCommand):
#     """Delete a stored ACP session.

#     This permanently removes the session from storage.
#     Use with caution as this action cannot be undone.
#     """

#     name = "delete-session"
#     category = "acp"

#     async def execute_command(
#         self,
#         ctx: CommandContext[ACPCommandContext],
#         session_id: str,
#         *,
#         confirm: bool = False,
#     ):
#         """Delete a stored ACP session.

#         Args:
#             ctx: Command context with ACP session
#             session_id: Session identifier to delete
#             confirm: Skip confirmation prompt
#         """
#         session = ctx.context.session
#         if not session.manager:
#             await ctx.print("❌ **Session manager not available**")
#             return

#         try:
#             if session.manager._persistent_manager:
#                 # Check if session exists
#                 session_data = (
#                     await session.manager._persistent_manager.load_session_data(
#                         session_id
#                     )
#                 )
#                 if not session_data:
#                     await ctx.print(f"❌ **Session not found:** `{session_id}`")
#                     return

#                 if not confirm:
#                     msg_count = len(session_data.conversation)
#                     await ctx.print(
#                         f"⚠️  **About to delete session `{session_id}`**"
#                     )
#                     await ctx.print(
#                         f"📊 **This session has {msg_count} messages**"
#                     )
#                     await ctx.print(
#                         f"**To confirm, run:** `/delete-session {session_id} --confirm`"
#                     )
#                     return

#                 # Delete the session
#                 await session.manager._persistent_manager.store.delete_session(session_id)  # noqa: E501
#                 await ctx.print(
#                     f"🗑️  **Session `{session_id}` deleted successfully**"
#                 )
#             else:
#                 await ctx.print("❌ **Session persistence not enabled**")

#         except Exception as e:
#             await ctx.print(f"❌ **Error deleting session:** {e}")


def get_acp_commands() -> list[type[NodeCommand]]:
    """Get all ACP-specific slash commands."""
    return [
        ListSessionsCommand,
        # LoadSessionCommand,
        # SaveSessionCommand,
        # DeleteSessionCommand,
    ]



"""Schema definitions for the ACP protocol."""

from acp.schema.agent_plan import PlanEntry, PlanEntryPriority, PlanEntryStatus
from acp.schema.agent_requests import (
    AgentRequest,
    CreateTerminalRequest,
    KillTerminalCommandRequest,
    ReadTextFileRequest,
    ReleaseTerminalRequest,
    RequestPermissionRequest,
    TerminalOutputRequest,
    WaitForTerminalExitRequest,
    WriteTextFileRequest,
)
from acp.schema.agent_responses import (
    AgentResponse,
    AuthenticateResponse,
    CustomResponse,
    InitializeResponse,
    LoadSessionResponse,
    NewSessionResponse,
    ListSessionsResponse,
    PromptResponse,
    SetSessionModeResponse,
    SetSessionModelResponse,
    StopReason,
)
from acp.schema.capabilities import (
    AgentCapabilities,
    ClientCapabilities,
    FileSystemCapability,
    McpCapabilities,
    PromptCapabilities,
    SessionCapabilities,
    SessionListCapabilities,
)
from acp.schema.client_requests import (
    AuthenticateRequest,
    ClientRequest,
    CustomRequest,
    InitializeRequest,
    ListSessionsRequest,
    LoadSessionRequest,
    NewSessionRequest,
    PromptRequest,
    SetSessionModeRequest,
    SetSessionModelRequest,
)
from acp.schema.client_responses import (
    ClientResponse,
    CreateTerminalResponse,
    KillTerminalCommandResponse,
    ReadTextFileResponse,
    ReleaseTerminalResponse,
    RequestPermissionResponse,
    TerminalOutputResponse,
    WaitForTerminalExitResponse,
    WriteTextFileResponse,
)
from acp.schema.common import AuthMethod, EnvVariable, Implementation
from acp.schema.content_blocks import (
    Annotations,
    AudioContentBlock,
    Audience,
    BlobResourceContents,
    ContentBlock,
    EmbeddedResourceContentBlock,
    ImageContentBlock,
    ResourceContentBlock,
    TextContentBlock,
    TextResourceContents,
)
from acp.schema.mcp import (
    HttpHeader,
    HttpMcpServer,
    McpServer,
    SseMcpServer,
    StdioMcpServer,
)
from acp.schema.messages import AgentMethod, ClientMethod
from acp.schema.notifications import (
    CancelNotification,
    ExtNotification,
    SessionNotification,
)
from acp.schema.session_state import (
    ModelInfo,
    SessionInfo,
    SessionMode,
    SessionModeState,
    SessionModelState,
)
from acp.schema.slash_commands import (
    AvailableCommand,
    AvailableCommandInput,
    CommandInputHint,
)
from acp.schema.terminal import TerminalExitStatus
from acp.schema.tool_call import (
    AllowedOutcome,
    ContentToolCallContent,
    DeniedOutcome,
    FileEditToolCallContent,
    PermissionKind,
    PermissionOption,
    TerminalToolCallContent,
    ToolCall,
    ToolCallContent,
    ToolCallKind,
    ToolCallLocation,
    ToolCallStatus,
)
from acp.schema.session_updates import (
    AgentMessageChunk,
    AgentPlanUpdate,
    AgentThoughtChunk,
    AvailableCommandsUpdate,
    CurrentModeUpdate,
    SessionUpdate,
    ToolCallProgress,
    ToolCallStart,
    UserMessageChunk,
)

PROTOCOL_VERSION = 1

__all__ = [
    # Protocol version
    "PROTOCOL_VERSION",
    # Agent capabilities
    "AgentCapabilities",
    # Session updates
    "AgentMessageChunk",
    # Messages/Methods
    "AgentMethod",
    "AgentPlanUpdate",
    # Agent requests (agent -> client)
    "AgentRequest",
    # Agent responses
    "AgentResponse",
    "AgentThoughtChunk",
    # Tool calls
    "AllowedOutcome",
    # Content blocks
    "Annotations",
    "Audience",
    "AudioContentBlock",
    # Common types
    "AuthMethod",
    # Client requests (client -> agent)
    "AuthenticateRequest",
    "AuthenticateResponse",
    # Slash commands
    "AvailableCommand",
    "AvailableCommandInput",
    "AvailableCommandsUpdate",
    "BlobResourceContents",
    # Notifications
    "CancelNotification",
    "ClientCapabilities",
    "ClientMethod",
    "ClientRequest",
    # Client responses
    "ClientResponse",
    "CommandInputHint",
    "ContentBlock",
    "ContentToolCallContent",
    "CreateTerminalRequest",
    "CreateTerminalResponse",
    "CurrentModeUpdate",
    "CustomRequest",
    "CustomResponse",
    "DeniedOutcome",
    "EmbeddedResourceContentBlock",
    "EnvVariable",
    "ExtNotification",
    "FileEditToolCallContent",
    "FileSystemCapability",
    # MCP servers
    "HttpHeader",
    "HttpMcpServer",
    "ImageContentBlock",
    "Implementation",
    "InitializeRequest",
    "InitializeResponse",
    "KillTerminalCommandRequest",
    "KillTerminalCommandResponse",
    "ListSessionsRequest",
    "ListSessionsResponse",
    "LoadSessionRequest",
    "LoadSessionResponse",
    "McpCapabilities",
    "McpServer",
    # Session state
    "ModelInfo",
    "NewSessionRequest",
    "NewSessionResponse",
    "PermissionKind",
    "PermissionOption",
    # Plan
    "PlanEntry",
    "PlanEntryPriority",
    "PlanEntryStatus",
    "PromptCapabilities",
    "PromptRequest",
    "PromptResponse",
    "ReadTextFileRequest",
    "ReadTextFileResponse",
    "ReleaseTerminalRequest",
    "ReleaseTerminalResponse",
    "RequestPermissionRequest",
    "RequestPermissionResponse",
    "ResourceContentBlock",
    "SessionCapabilities",
    "SessionInfo",
    "SessionListCapabilities",
    "SessionMode",
    "SessionModeState",
    "SessionModelState",
    "SessionNotification",
    "SessionUpdate",
    "SetSessionModeRequest",
    "SetSessionModeResponse",
    "SetSessionModelRequest",
    "SetSessionModelResponse",
    "SseMcpServer",
    "StdioMcpServer",
    "StopReason",
    # Terminal
    "TerminalExitStatus",
    "TerminalOutputRequest",
    "TerminalOutputResponse",
    "TerminalToolCallContent",
    "TextContentBlock",
    "TextResourceContents",
    "ToolCall",
    "ToolCallContent",
    "ToolCallKind",
    "ToolCallLocation",
    "ToolCallProgress",
    "ToolCallStart",
    "ToolCallStatus",
    "UserMessageChunk",
    "WaitForTerminalExitRequest",
    "WaitForTerminalExitResponse",
    "WriteTextFileRequest",
    "WriteTextFileResponse",
]



"""Search toolset implementation using searchly providers."""

from __future__ import annotations

from typing import TYPE_CHECKING

from llmling_agent.log import get_logger
from llmling_agent.resource_providers import ResourceProvider


if TYPE_CHECKING:
    from searchly.base import NewsSearchProvider, WebSearchProvider

    from llmling_agent.tools.base import Tool


logger = get_logger(__name__)


class SearchTools(ResourceProvider):
    """Provider for web and news search tools."""

    def __init__(
        self,
        web_search: WebSearchProvider | None = None,
        news_search: NewsSearchProvider | None = None,
    ) -> None:
        """Initialize search tools provider.

        Args:
            web_search: Web search provider instance.
            news_search: News search provider instance.
        """
        super().__init__(name="search")
        self._web_provider = web_search
        self._news_provider = news_search

    async def get_tools(self) -> list[Tool]:
        """Get search tools from configured providers."""
        tools: list[Tool] = []
        if self._web_provider:
            tools.append(
                self.create_tool(
                    self._web_provider.web_search,
                    read_only=True,
                    idempotent=True,
                    open_world=True,
                )
            )
        if self._news_provider:
            tools.append(
                self.create_tool(
                    self._news_provider.news_search,
                    read_only=True,
                    idempotent=True,
                    open_world=True,
                )
            )
        return tools



"""Fetch contents from a GitHub repository via UIThub."""

from __future__ import annotations

import os
import uuid

import httpx
from pydantic_ai import UserPromptPart
from slashed import CommandContext  # noqa: TC002

from llmling_agent.log import get_logger
from llmling_agent.messaging.context import NodeContext  # noqa: TC001
from llmling_agent_commands.base import NodeCommand
from llmling_agent_server.acp_server.session import ACPSession  # noqa: TC001


logger = get_logger(__name__)


class FetchRepoCommand(NodeCommand):
    """Fetch contents from a GitHub repository via UIThub.

    Retrieves repository contents with various filtering options
    and displays them in a structured format.
    """

    name = "fetch-repo"
    category = "docs"

    async def execute_command(  # noqa: PLR0915
        self,
        ctx: CommandContext[NodeContext[ACPSession]],
        repo: str,
        *,
        branch: str | None = None,
        path: str | None = None,
        include_dirs: list[str] | None = None,
        disable_genignore: bool = False,
        exclude_dirs: list[str] | None = None,
        exclude_extensions: list[str] | None = None,
        include_extensions: list[str] | None = None,
        include_line_numbers: bool = True,
        max_file_size: int | None = None,
        max_tokens: int | None = None,
        omit_files: bool = False,
        yaml_string: str | None = None,
    ) -> None:
        """Fetch contents from a GitHub repository.

        Args:
            ctx: Command context with ACP session
            repo: GitHub path (owner/repo)
            branch: Branch name (defaults to main if not provided)
            path: File or directory path within the repository
            include_dirs: List of directories to include
            disable_genignore: Disable .genignore filtering
            exclude_dirs: List of directories to exclude
            exclude_extensions: List of file extensions to exclude
            include_extensions: List of file extensions to include
            include_line_numbers: Include line numbers in HTML/markdown output
            max_file_size: Maximum file size in bytes
            max_tokens: Maximum number of tokens in response
            omit_files: Only return directory structure without file contents
            yaml_string: URL encoded YAML string of file hierarchy to include
        """
        session = ctx.context.data
        assert session

        # Generate tool call ID
        tool_call_id = f"fetch-repo-{uuid.uuid4().hex[:8]}"

        try:
            # Build URL
            base_url = f"https://uithub.com/{repo}"
            if branch:
                base_url += f"/tree/{branch}"
            if path:
                base_url += f"/{path}"

            # Build parameters
            params = {}
            api_key = os.getenv("UITHUB_API_KEY")
            if api_key:
                params["apiKey"] = api_key

            if include_dirs:
                params["dir"] = ",".join(include_dirs)
            if disable_genignore:
                params["disableGenignore"] = "true"
            if exclude_dirs:
                params["exclude-dir"] = ",".join(exclude_dirs)
            if exclude_extensions:
                params["exclude-ext"] = ",".join(exclude_extensions)
            if include_extensions:
                params["ext"] = ",".join(include_extensions)
            if not include_line_numbers:
                params["lines"] = "false"
            if max_file_size:
                params["maxFileSize"] = str(max_file_size)
            if max_tokens:
                params["maxTokens"] = str(max_tokens)
            if omit_files:
                params["omitFiles"] = "true"
            if yaml_string:
                params["yamlString"] = yaml_string

            # Start tool call
            display_path = f"{repo}"
            if branch:
                display_path += f"@{branch}"
            if path:
                display_path += f":{path}"

            await session.notifications.tool_call_start(
                tool_call_id=tool_call_id,
                title=f"Fetching repository: {display_path}",
                kind="fetch",
            )

            # Make async HTTP request
            async with httpx.AsyncClient() as client:
                response = await client.get(
                    base_url,
                    params=params,
                    headers={"accept": "text/markdown"},
                    timeout=30.0,
                )
                response.raise_for_status()
                content = response.text

            # Stage the content for use in agent context
            staged_part = UserPromptPart(
                content=f"Repository contents from {display_path}:\n\n{content}"
            )
            session.add_staged_parts([staged_part])

            # Send successful result - wrap in code block for proper display
            staged_count = session.get_staged_parts_count()
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="completed",
                title=f"Repository {display_path} fetched and staged ({staged_count} total parts)",
                content=[f"```\n{content}\n```"],
            )

        except httpx.HTTPStatusError as e:
            logger.exception(
                "HTTP error fetching repository", repo=repo, status=e.response.status_code
            )
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="failed",
                title=f"HTTP {e.response.status_code}: Failed to fetch {repo}",
            )
        except httpx.RequestError as e:
            logger.exception("Request error fetching repository", repo=repo)
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="failed",
                title=f"Network error: {e}",
            )
        except Exception as e:
            logger.exception("Unexpected error fetching repository", repo=repo)
            await session.notifications.tool_call_progress(
                tool_call_id=tool_call_id,
                status="failed",
                title=f"Error: {e}",
            )



"""Provider for history tools."""

from __future__ import annotations

from datetime import timedelta
from typing import Literal

from llmling_agent.agent.context import AgentContext  # noqa: TC001
from llmling_agent.resource_providers import StaticResourceProvider
from llmling_agent.utils.now import get_now


async def search_history(
    ctx: AgentContext,
    query: str | None = None,
    hours: int = 24,
    limit: int = 5,
) -> str:
    """Search conversation history."""
    from llmling_agent_storage.formatters import format_output

    if not ctx.pool:
        return "No agent pool available for history search"
    provider = ctx.pool.storage.get_history_provider()
    results = await provider.get_filtered_conversations(
        query=query,
        period=f"{hours}h",
        limit=limit,
    )
    return format_output(results)


async def show_statistics(
    ctx: AgentContext,
    group_by: Literal["agent", "model", "hour", "day"] = "model",
    hours: int = 24,
) -> str:
    """Show usage statistics for conversations."""
    from llmling_agent_storage.formatters import format_output
    from llmling_agent_storage.models import StatsFilters

    cutoff = get_now() - timedelta(hours=hours)
    filters = StatsFilters(cutoff=cutoff, group_by=group_by)

    if not ctx.pool:
        return "No agent pool available for statistics"
    provider = ctx.pool.storage.get_history_provider()
    stats = await provider.get_conversation_stats(filters)

    return format_output(
        {
            "period": f"{hours}h",
            "group_by": group_by,
            "entries": [
                {
                    "name": key,
                    "messages": data["messages"],
                    "total_tokens": data["total_tokens"],
                    "models": sorted(data["models"]),
                }
                for key, data in stats.items()
            ],
        },
        output_format="text",
    )


class HistoryTools(StaticResourceProvider):
    """Provider for history tools."""

    def __init__(self, name: str = "history") -> None:
        super().__init__(name=name)
        self._tools = [
            self.create_tool(search_history, category="search", read_only=True, idempotent=True),
            self.create_tool(show_statistics, category="read", read_only=True, idempotent=True),
        ]



"""Helper functions and constants for documentation commands."""

from __future__ import annotations


# Script constants for client-side execution
SCHEMA_EXTRACTION_SCRIPT = """
import sys
import json
import importlib

def import_callable(path):
    parts = path.split(".")
    for i in range(len(parts), 0, -1):
        try:
            module_path = ".".join(parts[:i])
            module = importlib.import_module(module_path)
            obj = module
            for part in parts[i:]:
                obj = getattr(obj, part)
            return obj
        except (ImportError, AttributeError):
            continue
    raise ImportError(f"Could not import %s" % path)

try:
    model_class = import_callable("{input_path}")
    schema = model_class.model_json_schema()
    print(json.dumps(schema, indent=2))
except Exception as e:
    print(f"Error: %s" % e, file=sys.stderr)
    sys.exit(1)
"""


def generate_from_url(input_path: str, class_name: str | None = None) -> str:
    """Generate Pydantic model code from an OpenAPI specification URL."""
    from datamodel_code_generator import DataModelType, LiteralType, PythonVersion
    from datamodel_code_generator.model import get_data_model_types
    from datamodel_code_generator.parser.openapi import OpenAPIParser

    data_model_types = get_data_model_types(
        DataModelType.PydanticV2BaseModel,
        target_python_version=PythonVersion.PY_312,
    )
    parser = OpenAPIParser(
        source=input_path,
        data_model_type=data_model_types.data_model,
        data_model_root_type=data_model_types.root_model,
        data_model_field_type=data_model_types.field_model,
        data_type_manager_type=data_model_types.data_type_manager,
        dump_resolve_reference_action=data_model_types.dump_resolve_reference_action,
        class_name=class_name,
        use_union_operator=True,
        use_schema_description=True,
        enum_field_as_literal=LiteralType.All,
    )
    return str(parser.parse())


def generate_from_schema(
    schema_json: str,
    class_name: str | None = None,
    input_path: str | None = None,
) -> str:
    """Generate Pydantic model code from a JSON schema."""
    from datamodel_code_generator import DataModelType, LiteralType, PythonVersion
    from datamodel_code_generator.model import get_data_model_types
    from datamodel_code_generator.parser.jsonschema import JsonSchemaParser

    data_model_types = get_data_model_types(
        DataModelType.PydanticV2BaseModel,
        target_python_version=PythonVersion.PY_312,
    )
    parser = JsonSchemaParser(
        source=schema_json,
        data_model_type=data_model_types.data_model,
        data_model_root_type=data_model_types.root_model,
        data_model_field_type=data_model_types.field_model,
        data_type_manager_type=data_model_types.data_type_manager,
        dump_resolve_reference_action=data_model_types.dump_resolve_reference_action,
        class_name=class_name or (input_path or "DefaultClass").split(".")[-1],
        use_union_operator=True,
        use_schema_description=True,
        enum_field_as_literal=LiteralType.All,
    )
    return str(parser.parse())



"""Prompt models for LLMling-Agent."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import field
import inspect
import os
from typing import TYPE_CHECKING, Annotated, Any, Literal, Self, assert_never

from fastmcp.prompts.prompt import (
    Prompt as FastMCPPrompt,
    PromptArgument as FastMCPArgument,
)
from mcp.types import Prompt as MCPPrompt, PromptArgument
from pydantic import ConfigDict, Field
from pydantic_ai import BinaryContent, SystemPromptPart, UserPromptPart
from pydantic_ai.messages import ImageUrl
from schemez import Schema
import upath
from upathtools import to_upath

from llmling_agent.log import get_logger
from llmling_agent.mcp_server import MCPClient
from llmling_agent.utils import importing
from llmling_agent.utils.inspection import execute


if TYPE_CHECKING:
    from collections.abc import Mapping

    from fastmcp.prompts.prompt import FunctionPrompt
    from pydantic_ai import ModelRequestPart


logger = get_logger(__name__)

MessageContentType = Literal["text", "resource", "image_url", "image_base64"]
# Our internal role type (could include more roles)
MessageRole = Literal["system", "user", "assistant", "tool"]

type CompletionFunction = Callable[[str], list[str]]
"""Type for completion functions. Takes current value, returns possible completions."""


class MessageContent(Schema):
    """Content item in a message."""

    type: MessageContentType
    content: str  # The actual content (text/uri/url/base64)
    alt_text: str | None = None  # For images or resource descriptions

    def to_pydantic(self) -> UserPromptPart:
        """Convert MessageContent to Pydantic model."""
        if self.type == "text":
            return UserPromptPart(self.content)
        if self.type == "image_url":
            return UserPromptPart([ImageUrl(self.content)])
        if self.type == "image_base64":
            bin_content = BinaryContent(self.content.encode(), media_type="image/jpeg")
            return UserPromptPart([bin_content])
        msg = "Unsupported content type"
        raise ValueError(msg)


class PromptParameter(Schema):
    """Prompt argument with validation information."""

    name: str
    """Name of the argument as used in the prompt."""

    description: str | None = None
    """Human-readable description of the argument."""

    required: bool = False
    """Whether this argument must be provided when formatting the prompt."""

    default: Any | None = None
    """Default value if argument is optional."""

    def to_mcp_argument(self) -> PromptArgument:
        """Convert to MCP PromptArgument."""
        from mcp.types import PromptArgument

        return PromptArgument(name=self.name, description=self.description, required=self.required)


class PromptMessage(Schema):
    """A message in a prompt template."""

    role: MessageRole
    content: str | MessageContent | list[MessageContent] = ""

    def get_text_content(self) -> str:
        """Get text content of message."""
        match self.content:
            case str():
                return self.content
            case MessageContent() if self.content.type == "text":
                return self.content.content
            case list() if self.content:
                # Join text content items with space
                text_items = [
                    item.content
                    for item in self.content
                    if isinstance(item, MessageContent) and item.type == "text"
                ]
                return " ".join(text_items) if text_items else ""
            case _:
                return ""

    def to_pydantic_parts(self) -> list[ModelRequestPart]:
        match self.role:
            case "system":
                return [SystemPromptPart(str(self.content))]
            case "user":
                match self.content:
                    case str():
                        return [UserPromptPart(self.content)]
                    case MessageContent():
                        return [self.content.to_pydantic()]
                    case list():
                        return [i.to_pydantic() for i in self.content]
        return []


class BasePrompt(Schema):
    """Base class for all prompts."""

    name: str | None = Field(None, exclude=True)
    """Technical identifier (automatically set from config key during registration)."""

    description: str
    """Human-readable description of what this prompt does."""

    title: str | None = None
    """Title of the prompt."""

    metadata: dict[str, Any] = Field(default_factory=dict)
    """Additional metadata for storing custom prompt information."""
    # messages: list[PromptMessage]

    async def format(self, arguments: dict[str, Any] | None = None) -> list[PromptMessage]:
        """Format this prompt with given arguments.

        Args:
            arguments: Optional argument values

        Returns:
            List of formatted messages

        Raises:
            ValueError: If required arguments are missing
        """
        raise NotImplementedError

    def to_mcp_prompt(self) -> MCPPrompt:
        """Convert to MCP Prompt."""
        raise NotImplementedError


class StaticPrompt(BasePrompt):
    """Static prompt defined by message list."""

    messages: list[PromptMessage]
    """List of messages that make up this prompt."""

    type: Literal["text"] = Field("text", init=False)
    """Discriminator field identifying this as a static text prompt."""

    arguments: list[PromptParameter] = Field(default_factory=list)
    """List of arguments that this prompt accepts."""

    def validate_arguments(self, provided: dict[str, Any]) -> None:
        """Validate that required arguments are provided."""
        required = {arg.name for arg in self.arguments if arg.required}
        missing = required - set(provided)
        if missing:
            msg = f"Missing required arguments: {', '.join(missing)}"
            raise ValueError(msg)

    def to_mcp_prompt(self) -> MCPPrompt:
        """Convert to MCP Prompt."""
        from mcp.types import Prompt as MCPPrompt

        if self.name is None:
            msg = "Prompt name not set. This should be set during registration."
            raise ValueError(msg)
        args = [arg.to_mcp_argument() for arg in self.arguments]
        return MCPPrompt(name=self.name, description=self.description, arguments=args)

    def to_fastmcp_prompt(self) -> FastMCPPrompt:
        params = [
            FastMCPArgument(name=p.name, description=p.description, required=p.required)
            for p in self.arguments
        ]
        return FastMCPPrompt(
            name=self.name or "",
            title=self.title,
            description=self.description,
            arguments=params,
            icons=None,
            tags=set(),
        )

    async def format(self, arguments: dict[str, Any] | None = None) -> list[PromptMessage]:
        """Format static prompt messages with arguments."""
        args = arguments or {}
        self.validate_arguments(args)

        # Add default values for optional arguments
        for arg in self.arguments:
            if arg.name not in args and not arg.required:
                args[arg.name] = arg.default if arg.default is not None else ""

        # Format all messages
        formatted_messages = []
        for msg in self.messages:
            match msg.content:
                case str():
                    content: MessageContent | list[MessageContent] = MessageContent(
                        type="text", content=msg.content.format(**args)
                    )
                case MessageContent() if msg.content.type == "text":
                    msg_content = msg.content.content.format(**args)
                    content = MessageContent(type="text", content=msg_content)
                case list():
                    content = [
                        MessageContent(
                            type=item.type,
                            content=item.content.format(**args)
                            if item.type == "text"
                            else item.content,
                            alt_text=item.alt_text,
                        )
                        for item in msg.content
                        if isinstance(item, MessageContent)
                    ]
                case _:
                    content = msg.content

            formatted_messages.append(PromptMessage(role=msg.role, content=content))

        return formatted_messages


class DynamicPrompt(BasePrompt):
    """Dynamic prompt loaded from callable."""

    import_path: str | Callable[..., str | Awaitable[str]]
    """Dotted import path to the callable that generates the prompt."""

    template: str | None = None
    """Optional template string for formatting the callable's output."""

    completions: dict[str, str] | None = None
    """Optional mapping of argument names to completion functions."""

    type: Literal["function"] = Field("function", init=False)
    """Discriminator field identifying this as a function-based prompt."""

    @property
    def messages(self) -> list[PromptMessage]:
        """Get the template messages for this prompt.

        Note: These are template messages - actual content will be populated
        during format() when the callable is executed.
        """
        template = self.template or "{result}"
        sys_content = MessageContent(type="text", content=f"Content from {self.name}:")
        user_content = MessageContent(type="text", content=template)
        return [
            PromptMessage(role="system", content=sys_content),
            PromptMessage(role="user", content=user_content),
        ]

    def to_fastmcp_prompt(self) -> FunctionPrompt:
        from fastmcp.prompts.prompt import FunctionPrompt

        return FunctionPrompt.from_function(self.fn, title=self.title, description=self.description)

    @property
    def fn(self) -> Callable[..., str | Awaitable[str]]:
        if isinstance(self.import_path, str):
            return importing.import_callable(self.import_path)
        return self.import_path

    def get_completion_functions(self) -> dict[str, CompletionFunction]:
        """Resolve completion function import paths and return a completion fn dict."""
        completion_funcs: dict[str, CompletionFunction] = {}
        if not self.completions:
            return {}
        for arg_name, import_path in self.completions.items():
            try:
                func = importing.import_callable(import_path)
                completion_funcs[arg_name] = func
            except ValueError:
                msg = "Failed to import completion function for %s: %s"
                logger.warning(msg, arg_name, import_path)
        return completion_funcs

    async def format(self, arguments: dict[str, Any] | None = None) -> list[PromptMessage]:
        """Format this prompt with given arguments."""
        args = arguments or {}
        try:
            result: Any = await execute(self.fn, **args)
            template = self.template or "{result}"
            msg = template.format(result=result)
            content = MessageContent(type="text", content=msg)
            msg = f"Content from {self.name}:"
            sys_content = MessageContent(type="text", content=msg)
            return [
                PromptMessage(role="system", content=sys_content),
                PromptMessage(role="user", content=content),
            ]
        except Exception as exc:
            msg = f"Failed to execute prompt callable: {exc}"
            raise ValueError(msg) from exc

    @classmethod
    def from_callable(
        cls,
        fn: Callable[..., Any] | str,
        *,
        name_override: str | None = None,
        description_override: str | None = None,
        template_override: str | None = None,
        completions: Mapping[str, CompletionFunction] | None = None,
    ) -> DynamicPrompt:
        """Create a prompt from a callable.

        Args:
            fn: Function or import path to create prompt from
            name_override: Optional override for prompt name
            description_override: Optional override for prompt description
            template_override: Optional override for message template
            completions: Optional dict mapping argument names to completion functions

        Returns:
            DynamicPrompt instance

        Raises:
            ValueError: If callable cannot be imported or is invalid
        """
        from docstring_parser import parse as parse_docstring

        completions = completions or {}
        if isinstance(fn, str):
            fn = importing.import_callable(fn)

        name = name_override or getattr(fn, "__name__", "unknown")
        if docstring := inspect.getdoc(fn):
            parsed = parse_docstring(docstring)
            description = description_override or parsed.short_description
        else:
            description = description_override or f"Prompt from {name}"

        path = f"{fn.__module__}.{fn.__qualname__}"
        return cls(
            name=name,
            description=description or "",
            import_path=path,
            template=template_override,
            metadata={"source": "function", "import_path": path},
        )

    def to_mcp_prompt(self) -> MCPPrompt:
        """Convert to MCP Prompt."""
        from docstring_parser import parse as parse_docstring
        from mcp.types import Prompt as MCPPrompt

        sig = inspect.signature(self.fn)
        # hints = get_type_hints(self.fn, include_extras=True, localns=locals())
        if docstring := inspect.getdoc(self.fn):
            parsed = parse_docstring(docstring)
            # Create mapping of param names to descriptions
            arg_docs = {
                param.arg_name: param.description
                for param in parsed.params
                if param.arg_name and param.description
            }
        else:
            arg_docs = {}

        # Create arguments
        arguments = []
        for param_name, param in sig.parameters.items():
            if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
                continue

            required = param.default == param.empty
            arg = PromptParameter(
                name=param_name,
                description=arg_docs.get(param_name),
                required=required,
                default=None if param.default is param.empty else param.default,
            )
            arguments.append(arg)

        if self.name is None:
            msg = "Prompt name not set. This should be set during registration."
            raise ValueError(msg)
        args = [arg.to_mcp_argument() for arg in arguments]
        return MCPPrompt(name=self.name, description=self.description, arguments=args)


class MCPClientPrompt(BasePrompt):
    """A prompt that can be rendered from an MCP server."""

    client: MCPClient
    """The client used to render the prompt."""

    name: str
    """The name of the prompt."""

    arguments: list[dict[str, Any]] = field(default_factory=list)
    """A list of arguments for the prompt."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_mcp_prompt(self) -> MCPPrompt:
        """Convert to MCP Prompt."""
        args = [
            PromptArgument(name=i["name"], description=i["description"], required=i["required"])
            for i in self.arguments
        ]
        return MCPPrompt(name=self.name, description=self.description, arguments=args)

    @classmethod
    def from_fastmcp(cls, client: MCPClient, prompt: MCPPrompt) -> Self:
        """Convert MCP prompt to our Prompt class."""
        arguments = [
            {
                "name": arg.name,
                "description": arg.description,
                "required": arg.required or False,
            }
            for arg in prompt.arguments or []
        ]

        return cls(
            name=prompt.name,
            description=prompt.description or "",
            arguments=arguments,
            client=client,
        )

    def __repr__(self) -> str:
        return f"Prompt(name={self.name!r}, description={self.description!r})"

    async def get_components(
        self, arguments: dict[str, str] | None = None
    ) -> list[SystemPromptPart | UserPromptPart]:
        """Get prompt as pydantic-ai message components.

        Args:
            arguments: Arguments to pass to the prompt template

        Returns:
            List of message parts ready for agent usage

        Raises:
            RuntimeError: If prompt fetch fails
            ValueError: If prompt contains unsupported message types
        """
        from llmling_agent.mcp_server.conversions import content_block_as_text

        try:
            result = await self.client.get_prompt(self.name, arguments)
        except Exception as e:
            msg = f"Failed to get prompt {self.name!r}: {e}"
            raise RuntimeError(msg) from e

        # Convert MCP messages to pydantic-ai parts
        parts: list[SystemPromptPart | UserPromptPart] = []
        for message in result.messages:
            # Extract text content from MCP message
            text_content = content_block_as_text(message.content)
            # Convert based on role
            match message.role:
                case "user":
                    parts.append(UserPromptPart(content=text_content))
                case "assistant":
                    # Convert assistant messages to user parts for context
                    parts.append(UserPromptPart(content=f"Assistant: {text_content}"))
                case _ as unreachable:
                    assert_never(unreachable)

        if not parts:
            msg = f"No supported message parts found in prompt {self.name!r}"
            raise ValueError(msg)

        return parts


class FilePrompt(BasePrompt):
    """Prompt loaded from a file.

    This type of prompt loads its content from a file, allowing for longer or more
    complex prompts to be managed in separate files. The file content is loaded
    and parsed according to the specified format.
    """

    path: str | os.PathLike[str] | upath.UPath
    """Path to the file containing the prompt content."""

    fmt: Literal["text", "markdown", "jinja2"] = Field("text", alias="format")
    """Format of the file content (text, markdown, or jinja2 template)."""

    type: Literal["file"] = Field("file", init=False)
    """Discriminator field identifying this as a file-based prompt."""

    watch: bool = False
    """Whether to watch the file for changes and reload automatically."""

    def to_mcp_prompt(self) -> MCPPrompt:
        """Convert to MCP Prompt."""
        return MCPPrompt(
            name=self.name or upath.UPath(self.path).name,
            description=self.description,
        )

    @property
    def messages(self) -> list[PromptMessage]:
        """Get messages from file content."""
        content = to_upath(self.path).read_text("utf-8")
        match self.fmt:
            case "text":
                # Simple text format - whole file as user message
                msg = MessageContent(type="text", content=content)
            case "markdown":
                # TODO: Parse markdown sections into separate messages
                msg = MessageContent(type="text", content=content)
            case "jinja2":
                # Raw template - will be formatted during format()
                msg = MessageContent(type="text", content=content)
            case _ as unreachable:
                assert_never(unreachable)
        return [PromptMessage(role="user", content=msg)]

    async def format(self, arguments: dict[str, Any] | None = None) -> list[PromptMessage]:
        """Format the file content with arguments."""
        args = arguments or {}
        content = to_upath(self.path).read_text("utf-8")

        if self.fmt == "jinja2":
            # Use jinja2 for template formatting
            import jinja2

            env = jinja2.Environment(autoescape=True, enable_async=True)
            template = env.from_string(content)
            content = await template.render_async(**args)
        else:
            # Use simple string formatting
            try:
                content = content.format(**args)
            except KeyError as exc:
                msg = f"Missing argument in template: {exc}"
                raise ValueError(msg) from exc
        msg_content = MessageContent(type="text", content=content)
        return [PromptMessage(role="user", content=msg_content)]


# Type to use in configuration
PromptType = Annotated[StaticPrompt | DynamicPrompt | FilePrompt, Field(discriminator="type")]


if __name__ == "__main__":

    def prompt_fn() -> str:
        return "hello"

    async def main() -> None:
        prompt = DynamicPrompt(import_path=prompt_fn, name="test", description="test")
        result = await prompt.format()
        print(result)

    import asyncio

    asyncio.run(main())



"""Provider for agent pool building tools."""

from __future__ import annotations

import asyncio
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Literal

from pydantic_ai import ModelRetry

from llmling_agent.agent.context import AgentContext  # noqa: TC001
from llmling_agent.log import get_logger
from llmling_agent.resource_providers import StaticResourceProvider
from llmling_agent.tools.exceptions import ToolError
from llmling_agent.utils.result_utils import to_type


if TYPE_CHECKING:
    from llmling_agent.agent import Agent

logger = get_logger(__name__)


async def create_worker_agent[TDeps](
    ctx: AgentContext[TDeps],
    name: str,
    system_prompt: str,
    model: str | None = None,
) -> str:
    """Create a new agent and register it as a tool.

    The new agent will be available as a tool for delegating specific tasks.
    It inherits the current model unless overridden.
    """
    from llmling_agent import Agent

    if not ctx.pool:
        msg = "Agent needs to be in a pool to list agents"
        raise ToolError(msg)

    model = model or ctx.agent.model_name
    agent = Agent[TDeps](name=name, model=model, system_prompt=system_prompt, agent_pool=ctx.pool)
    assert ctx.agent
    tool_info = ctx.agent.register_worker(agent)
    return f"Created worker agent and registered as tool: {tool_info.name}"


async def add_agent(  # noqa: D417
    ctx: AgentContext,
    name: str,
    system_prompt: str,
    model: str | None = None,
    tools: list[str] | None = None,
    session: str | None = None,
    output_type: str | None = None,
) -> str:
    """Add a new agent to the pool.

    Args:
        name: Name for the new agent
        system_prompt: System prompt defining agent's role/behavior
        model: Optional model override (uses default if not specified)
        tools: Imort paths of the tools the agent should have, if any.
        session: Session ID to recover conversation state from
        output_type: Name of response type from manifest (for structured output)

    Returns:
        Confirmation message about the created agent
    """
    assert ctx.pool, "No agent pool available"
    try:
        agent: Agent[Any, Any] = await ctx.pool.add_agent(
            name=name,
            system_prompt=system_prompt,
            model=model,
            tools=tools,
            output_type=to_type(output_type, responses=ctx.pool.manifest.responses),
            session=session,
        )
    except ValueError as e:  # for wrong tool imports
        raise ModelRetry(message=f"Error creating agent: {e}") from None
    return f"Created agent {agent.name} using model {agent.model_name}"


async def add_team(  # noqa: D417
    ctx: AgentContext,
    nodes: list[str],
    mode: Literal["sequential", "parallel"] = "sequential",
    name: str | None = None,
) -> str:
    """Create a team from existing agents.

    Args:
        nodes: Names of agents / sub-teams to include in team
        mode: How the team should operate:
            - sequential: Agents process in sequence (pipeline)
            - parallel: Agents process simultaneously
        name: Optional name for the team
    """
    if not ctx.pool:
        msg = "No agent pool available"
        raise ToolError(msg)

    # Verify all agents exist
    for node_name in nodes:
        if node_name not in ctx.pool.nodes:
            msg = (
                f"No agent or team found with name: {node_name}. "
                f"Available nodes: {', '.join(ctx.pool.nodes.keys())}"
            )
            raise ModelRetry(msg)
    if mode == "sequential":
        ctx.pool.create_team_run(nodes, name=name)
    else:
        ctx.pool.create_team(nodes, name=name)
    mode_str = "pipeline" if mode == "sequential" else "parallel"
    return f"Created {mode_str} team with nodes: {', '.join(nodes)}"


async def connect_nodes(  # noqa: D417
    ctx: AgentContext,
    source: str,
    target: str,
    *,
    connection_type: Literal["run", "context", "forward"] = "run",
    priority: int = 0,
    delay_seconds: float | None = None,
    queued: bool = False,
    queue_strategy: Literal["concat", "latest", "buffer"] = "latest",
    wait_for_completion: bool = True,
    name: str | None = None,
) -> str:
    """Connect two nodes to enable message flow between them.

    Nodes can be agents or teams.

    Args:
        source: Name of the source node
        target: Name of the target node
        connection_type: How messages should be handled:
            - run: Execute message as a new run in target
            - context: Add message as context to target
            - forward: Forward message to target's outbox
        priority: Task priority (lower = higher priority)
        delay_seconds: Optional delay before processing messages
        queued: Whether messages should be queued for manual processing
        queue_strategy: How to process queued messages:
            - concat: Combine all messages with newlines
            - latest: Use only the most recent message
            - buffer: Process all messages individually
        wait_for_completion: Whether to wait for target to complete
        name: Optional name for this connection

    Returns:
        Description of the created connection
    """
    if not ctx.pool:
        msg = "No agent pool available"
        raise ToolError(msg)

    # Get the nodes
    if source not in ctx.pool.nodes:
        msg = (
            f"No agent or team found with name: {source}. "
            f"Available nodes: {', '.join(ctx.pool.nodes.keys())}"
        )
        raise ModelRetry(msg)
    if target not in ctx.pool.nodes:
        msg = (
            f"No agent or team found with name: {target}. "
            f"Available nodes: {', '.join(ctx.pool.nodes.keys())}"
        )
        raise ModelRetry(msg)

    source_node = ctx.pool.nodes[source]
    target_node = ctx.pool.nodes[target]

    # Create the connection
    delay = timedelta(seconds=delay_seconds) if delay_seconds is not None else None
    _talk = source_node.connect_to(
        target_node,
        connection_type=connection_type,
        priority=priority,
        delay=delay,
        queued=queued,
        queue_strategy=queue_strategy,
        name=name,
    )
    source_node.connections.set_wait_state(target_node, wait=wait_for_completion)

    return (
        f"Created connection from {source} to {target} "
        f"(type={connection_type}, queued={queued}, "
        f"strategy={queue_strategy if queued else 'n/a'})"
    )


class AgentManagementTools(StaticResourceProvider):
    """Provider for agent pool building tools."""

    def __init__(self, name: str = "agent_management") -> None:
        super().__init__(name=name)
        for tool in [
            self.create_tool(create_worker_agent, category="other"),
            self.create_tool(add_agent, category="other"),
            self.create_tool(add_team, category="other"),
            self.create_tool(connect_nodes, category="other"),
        ]:
            self.add_tool(tool)


if __name__ == "__main__":
    # import logging
    from llmling_agent import AgentPool

    user_prompt = """Add a stdio MCP server:
// 	"command": "npx",
// 	"args": ["mcp-graphql"],
// 	"env": { "ENDPOINT": "https://diego.one/graphql" }

."""

    async def main() -> None:
        from llmling_agent_config.toolsets import IntegrationToolsetConfig

        async with AgentPool() as pool:
            toolsets = [IntegrationToolsetConfig()]
            toolset_providers = [config.get_provider() for config in toolsets]
            agent = await pool.add_agent(
                "X",
                toolsets=toolset_providers,
                model="openai:gpt-5-nano",
            )
            result = await agent.run(user_prompt)
            print(result)
            result = await agent.run("Which tools does it have?")
            print(result)

    asyncio.run(main())



"""Task Supervisor."""

from __future__ import annotations

import asyncio
from collections.abc import Callable
from contextlib import suppress
from typing import TYPE_CHECKING, Any

from llmling_agent import log


if TYPE_CHECKING:
    from collections.abc import Coroutine


__all__ = ["TaskSupervisor"]

logger = log.get_logger(__name__)
ErrorHandler = Callable[[asyncio.Task[Any], BaseException], None]


class TaskSupervisor:
    """Track background tasks and provide graceful shutdown semantics.

    Inspired by fasta2a's task manager, this supervisor keeps a registry of
    asyncio tasks created for request handling so they can be cancelled and
    awaited reliably when the connection closes.
    """

    def __init__(self, *, source: str) -> None:
        self._source = source
        self._tasks: set[asyncio.Task[Any]] = set()
        self._closed = False
        self._error_handlers: list[ErrorHandler] = []

    def add_error_handler(self, handler: ErrorHandler) -> None:
        self._error_handlers.append(handler)

    def create(
        self,
        coroutine: Coroutine[Any, Any, Any],
        *,
        name: str | None = None,
        on_error: ErrorHandler | None = None,
    ) -> asyncio.Task[Any]:
        if self._closed:
            msg = f"TaskSupervisor for {self._source} already closed"
            raise RuntimeError(msg)
        task = asyncio.create_task(coroutine, name=name)
        self._tasks.add(task)
        task.add_done_callback(lambda t: self._on_done(t, on_error))
        return task

    def _on_done(self, task: asyncio.Task[Any], on_error: ErrorHandler | None) -> None:
        self._tasks.discard(task)
        if task.cancelled():
            return
        try:
            task.result()
        except Exception as exc:
            handled = False
            if on_error is not None:
                try:
                    on_error(task, exc)
                    handled = True
                except Exception:
                    logger.exception(
                        "Error in task-specific error handler",
                        source=self._source,
                    )
            if not handled:
                for handler in self._error_handlers:
                    try:
                        handler(task, exc)
                        handled = True
                    except Exception:
                        msg = "Error in supervisor error handler"
                        logger.exception(msg, source=self._source)
            if not handled:
                logger.exception("Unhandled error in task", source=self._source)

    async def shutdown(self) -> None:
        self._closed = True
        if not self._tasks:
            return
        tasks = list(self._tasks)
        for task in tasks:
            task.cancel()
        for task in tasks:
            with suppress(asyncio.CancelledError):
                await task
        self._tasks.clear()



"""Prepare LLM output to be in proper shape and executable."""

from __future__ import annotations

import ast
from typing import TYPE_CHECKING

from llmling_agent.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Sequence

    from schemez import ToolsetCodeGenerator

    from llmling_agent.tools import Tool

logger = get_logger(__name__)


def validate_code(python_code: str) -> None:
    """Validate code structure and raise ModelRetry for fixable issues."""
    from pydantic_ai import ModelRetry

    code = python_code.strip()
    try:
        ast.parse(code)
    except SyntaxError as e:
        msg = f"Invalid code syntax: {e}"
        logger.info("Invalid code", code=code)
        raise ModelRetry(msg) from None
    else:
        if "async def main(" not in code:
            logger.info("Code not in async def main()", code=code)
            msg = (
                "Code must be wrapped in 'async def main():' function. "
                "Please rewrite your code like:\n"
                "async def main():\n"
                "    # your code here\n"
                "    return result"
            )
            raise ModelRetry(msg)

        # Check if code contains a return statement
        if "return " not in code:
            logger.info("Code does not contain a return statement", code=code)
            msg = "The main() function should return a value."
            raise ModelRetry(msg)


def tools_to_codegen(
    tools: Sequence[Tool],
    include_docstrings: bool = True,
) -> ToolsetCodeGenerator:
    """Create a ToolsetCodeGenerator from a sequence of Tools.

    Args:
        tools: Tools to generate code for
        include_docstrings: Include function docstrings in documentation

    Returns:
        ToolsetCodeGenerator instance
    """
    from pydantic_ai import RunContext
    from schemez import ToolCodeGenerator, ToolsetCodeGenerator, create_schema

    from llmling_agent.agent.context import AgentContext

    generators = [
        ToolCodeGenerator(
            schema=create_schema(
                t.callable,
                name_override=t.name,
                description_override=t.description,
                mode="openai",
                exclude_types=[AgentContext, RunContext],
            ),
            callable=t.callable,
            name_override=t.name,
        )
        for t in tools
    ]
    return ToolsetCodeGenerator(generators, include_docstrings)



"""Team configuration models."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from evented.configs import EventConfig
from pydantic import ConfigDict, Field, ImportString
from schemez import Schema

from llmling_agent_config.event_handlers import EventHandlerConfig
from llmling_agent_config.forward_targets import ForwardingTarget
from llmling_agent_config.mcp_server import (
    BaseMCPServerConfig,
    MCPServerConfig,
    StdioMCPServerConfig,
)


if TYPE_CHECKING:
    from llmling_agent.common_types import IndividualEventHandler


class NodeConfig(Schema):
    """Configuration for a Node of the messaging system."""

    model_config = ConfigDict(
        frozen=True,
        arbitrary_types_allowed=True,
        json_schema_extra={
            "x-icon": "octicon:workflow-16",
            "x-doc-title": "Node Configuration",
        },
    )

    name: str | None = Field(default=None)
    """Identifier for the node. Set from dict key, not from YAML."""

    config_file_path: str | None = Field(
        default=None,
        examples=["/path/to/config.yml", "configs/agent.yaml"],
        title="Configuration file path",
    )
    """Config file path for resolving relative paths."""

    display_name: str | None = Field(
        default=None,
        examples=["Main Agent", "Web Searcher", "Code Assistant"],
        title="Display name",
    )
    """Human-readable display name for the node."""

    description: str | None = Field(
        default=None,
        examples=["Main conversation agent", "Handles web search requests"],
        title="Node description",
    )
    """Optional description of the agent / team."""

    triggers: list[EventConfig] = Field(
        default_factory=list,
        examples=[
            [
                {
                    "type": "time",
                    "name": "daily_check",
                    "schedule": "0 9 * * *",
                    "prompt": "Daily status update",
                }
            ],
            [
                {
                    "type": "file",
                    "name": "code_watcher",
                    "paths": ["./src"],
                    "extensions": [".py"],
                }
            ],
        ],
        title="Event triggers",
    )
    """Event sources that activate this agent / team"""

    connections: list[ForwardingTarget] = Field(
        default_factory=list,
        examples=[
            [
                {
                    "type": "node",
                    "name": "output_agent",
                    "connection_type": "run",
                    "wait_for_completion": True,
                }
            ],
            [
                {
                    "type": "file",
                    "path": "logs/messages.txt",
                    "template": "{{ message.content }}",
                }
            ],
        ],
        title="Message forwarding targets",
    )
    """Targets to forward results to."""

    mcp_servers: list[str | MCPServerConfig] = Field(
        default_factory=list,
        title="MCP servers",
        examples=[
            ["uvx some-server"],
            [{"type": "streamable-http", "url": "http://mcp.example.com"}],
        ],
    )
    """List of MCP server configurations:
    - str entries are converted to StdioMCPServerConfig
    - MCPServerConfig for full server configuration
    """
    # Any should be InputProvider, but this leads to circular import
    input_provider: ImportString[Any] | None = Field(
        default=None,
        title="Input provider",
    )
    """Provider for human-input-handling."""

    event_handlers: list[EventHandlerConfig] = Field(
        default_factory=list,
        title="Event handlers",
        examples=[
            [{"type": "builtin", "handler": "simple"}],
        ],
    )
    """Event handlers for processing agent stream events.

    Supports:
    - builtin: Simple/detailed console output
    - tts: Text-to-speech synthesis
    - callback: Custom handler via import path
    """

    def get_event_handlers(self) -> list[IndividualEventHandler]:
        """Get resolved event handlers from configuration.

        Returns:
            List of event handler callables.
        """
        from llmling_agent_config.event_handlers import resolve_handler_configs

        return resolve_handler_configs(self.event_handlers)

    def get_mcp_servers(self) -> list[MCPServerConfig]:
        """Get processed MCP server configurations.

        Converts string entries to StdioMCPServerConfigs by splitting
        into command and arguments.

        Returns:
            List of MCPServerConfig instances

        Raises:
            ValueError: If string entry is empty
        """
        configs: list[MCPServerConfig] = []

        for server in self.mcp_servers:
            match server:
                case str():
                    parts = server.split()
                    if not parts:
                        msg = "Empty MCP server command"
                        raise ValueError(msg)

                    configs.append(StdioMCPServerConfig(command=parts[0], args=parts[1:]))
                case BaseMCPServerConfig():
                    configs.append(server)

        return configs
